from .convvt import ConvEmbedding
from .cvt import CVTEmbedding
from .linear import LinearEmbedding
from .overlappatch import OverlapPatchEmbed
from .patch import PatchEmbedding
from .pos_embedding import *
from .video_patch_embeddings import *



import torch.nn as nn

from ...utils import pair


class OverlapPatchEmbed(nn.Module):
    """

    Parameters
    ----------
    img_size: int
        Image Size
    patch_size: int or tuple(int)
        Patch Size
    stride: int
        Stride of the convolution, default is 4
    in_channels: int
        Number of input channels in the image, default is 3
    embedding_dim: int
        Number of linear projection output channels,default is 768
    norm_layer: nn.Module, optional
        Normalization layer, default is nn.LayerNorm

    """

    def __init__(
        self,
        img_size,
        patch_size,
        stride=4,
        in_channels=3,
        embedding_dim=768,
        norm_layer=nn.LayerNorm,
    ):
        super(OverlapPatchEmbed, self).__init__()

        img_size = pair(img_size)
        patch_size = pair(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size

        self.H, self.W = img_size[0] // stride, img_size[1] // stride

        self.proj = nn.Conv2d(
            in_channels=in_channels,
            out_channels=embedding_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=(patch_size[0] // 2, patch_size[1] // 2),
        )
        self.norm = norm_layer(embedding_dim)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
              Input tensor

        Returns
        ----------
        x: torch.Tensor
            Input tensor
        H: int
            Height of Patch
        W: int
            Width of Patch

        """

        x = self.proj(x)
        H, W = x.shape[2:]
        x = self.norm(x.flatten(2).transpose(1, 2))

        return x, H, W



from .convit import ConViTEncoder
from .convvt import ConvVTStage
from .cross import CrossEncoder
from .embedding import *
from .nn import FeedForward
from .pyramid import PVTEncoder
from .swin import SwinEncoder, SwinEncoderBlock
from .vanilla import VanillaEncoder
from .vivit import ViViTEncoder, ViViTEncoderBlock



import torch.nn as nn
from timm.models.layers import DropPath

from ..attention import SpatialAttention
from ..common.blocks import DWConv
from ..functional import PreNorm
from ..utils import ENCODER_REGISTRY


class PVTFeedForward(nn.Module):
    """

    Parameters
    ----------
    dim: int
        Dimension of the input tensor
    hidden_dim: int, optional
        Dimension of hidden layer
    out_dim:int, optional
        Dimension of output tensor
    act_layer: nn.Module
        Activation Layer, default is nn.GELU
    p_dropout: float
        Dropout probability/rate, default is 0.0
    linear: bool
        Whether to use linear Spatial attention,default is False
    use_dwconv: bool
        Whether to use Depth-wise convolutions, default is False

    kernel_size_dwconv: int
        `kernel_size` parameter for 2D convolution used in Depth wise convolution
    stride_dwconv: int
        `stride` parameter for 2D convolution used in Depth wise convolution
    padding_dwconv: int
        `padding` parameter for 2D convolution used in Depth wise convolution
    bias_dwconv:bool
        `bias` parameter for 2D convolution used in Depth wise convolution
    """

    def __init__(
        self,
        dim,
        hidden_dim=None,
        out_dim=None,
        act_layer=nn.GELU,
        p_dropout=0.0,
        linear=False,
        use_dwconv=False,
        **kwargs
    ):
        super(PVTFeedForward, self).__init__()

        out_dim = out_dim if out_dim is not None else dim
        hidden_dim = hidden_dim if hidden_dim is not None else dim
        self.use_dwconv = use_dwconv
        self.fc1 = nn.Linear(dim, hidden_dim)
        self.relu = nn.ReLU(inplace=True) if linear else nn.Identity()

        if use_dwconv:
            self.dw_conv = DWConv(dim=hidden_dim, **kwargs)

        self.to_out = nn.Sequential(
            act_layer(),
            nn.Dropout(p=p_dropout),
            nn.Linear(hidden_dim, out_dim),
            nn.Dropout(p=p_dropout),
        )

    def forward(self, x, **kwargs):

        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        H: int
            Height of image patch
        W: int
            Width of image patch

        Returns
        --------
        torch.Tensor
            Returns output tensor

        """
        x = self.relu(self.fc1(x))

        if self.use_dwconv:
            x = self.dw_conv(x, **kwargs)

        return self.to_out(x)


@ENCODER_REGISTRY.register()
class PVTEncoder(nn.Module):
    """
    Parameters
    ----------
    dim: int
        Dimension of the input tensor
    num_heads: int
        Number of attention heads
    mlp_ratio:
        Ratio of MLP hidden dimension to embedding dimension
    depth: int
        Number of attention layers in the encoder
    qkv_bias: bool
        Whether to add a bias vector to the q,k, and v matrices
    qk_scale:float, optional
        Override default qk scale of head_dim ** -0.5 in Spatial Attention if set
    p_dropout: float
        Dropout probability
    attn_dropout: float
        Dropout probability
    drop_path: tuple(float)
        List of stochastic drop rate
    act_layer: activation layer
        Activation layer
    use_dwconv:bool
        Whether to use depth-wise convolutions in overlap-patch embedding
    sr_ratio: float
        Spatial Reduction ratio
    linear: bool
        Whether to use linear Spatial attention, default is False
    """

    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio,
        depth,
        qkv_bias,
        qk_scale,
        p_dropout,
        attn_dropout,
        drop_path,
        act_layer,
        use_dwconv,
        sr_ratio,
        linear=False,
    ):
        super(PVTEncoder, self).__init__()

        self.encoder = nn.ModuleList([])

        for i in range(depth):
            self.encoder.append(
                nn.ModuleList(
                    [
                        PreNorm(
                            dim=dim,
                            fn=SpatialAttention(
                                dim=dim,
                                num_heads=num_heads,
                                qkv_bias=qkv_bias,
                                qk_scale=qk_scale,
                                attn_drop=attn_dropout,
                                proj_drop=p_dropout,
                                sr_ratio=sr_ratio,
                                linear=linear,
                            ),
                        ),
                        PreNorm(
                            dim=dim,
                            fn=PVTFeedForward(
                                dim=dim,
                                hidden_dim=int(dim * mlp_ratio),
                                act_layer=act_layer,
                                p_dropout=p_dropout,
                                linear=linear,
                                use_dwconv=use_dwconv,
                            ),
                        ),
                    ]
                )
            )
            self.drop_path = (
                DropPath(drop_prob=drop_path[i])
                if drop_path[i] > 0.0
                else nn.Identity()
            )

    def forward(self, x, **kwargs):

        for prenorm_attn, prenorm_ff in self.encoder:
            x = x + self.drop_path(prenorm_attn(x, **kwargs))
            x = x + self.drop_path(prenorm_ff(x, **kwargs))

        return x



import torch

from vformer.utils import ATTENTION_REGISTRY

attention_modules = ATTENTION_REGISTRY.get_list()


def test_VanillaSelfAttention():

    test_tensor1 = torch.randn(2, 65, 1024)
    test_tensor2 = torch.randn(2, 257, 1024)

    attention = ATTENTION_REGISTRY.get("VanillaSelfAttention")(dim=1024)
    out = attention(test_tensor1)
    assert out.shape == (2, 65, 1024)
    del attention

    attention = ATTENTION_REGISTRY.get("VanillaSelfAttention")(dim=1024, num_heads=16)
    out = attention(test_tensor2)
    assert out.shape == (2, 257, 1024)
    del attention


def test_WindowAttention():

    test_tensor1 = torch.randn(256, 49, 96)
    test_tensor2 = torch.randn(32, 64, 96)

    attention = ATTENTION_REGISTRY.get("WindowAttention")(
        dim=96, window_size=7, num_heads=3
    )
    out = attention(test_tensor1)
    assert out.shape == test_tensor1.shape
    del attention

    attention = ATTENTION_REGISTRY.get("WindowAttention")(
        dim=96, window_size=8, num_heads=4
    )
    out = attention(test_tensor2)
    assert out.shape == test_tensor2.shape
    del attention


def test_CrossAttention():

    test_tensor1 = torch.randn(64, 1, 64)
    test_tensor2 = torch.randn(64, 24, 128)

    attention = ATTENTION_REGISTRY.get("CrossAttention")(64, 128, 64)
    out = attention(test_tensor1, test_tensor2)
    assert out.shape == test_tensor1.shape
    del attention


def test_SpatialAttention():

    test_tensor1 = torch.randn(4, 3136, 64)
    test_tensor2 = torch.randn(4, 50, 512)

    attention = ATTENTION_REGISTRY.get("SpatialAttention")(
        dim=64,
        num_heads=1,
        sr_ratio=8,
    )
    out = attention(test_tensor1, H=56, W=56)
    assert out.shape == test_tensor1.shape
    del attention

    attention = ATTENTION_REGISTRY.get("SpatialAttention")(
        dim=512, num_heads=8, sr_ratio=1, linear=False
    )
    out = attention(test_tensor2, H=7, W=7)
    assert out.shape == test_tensor2.shape
    del attention

    attention = ATTENTION_REGISTRY.get("SpatialAttention")(
        dim=64, num_heads=1, sr_ratio=8, linear=True
    )
    out = attention(test_tensor1, 56, 56)
    assert out.shape == test_tensor1.shape


def test_GatedPositionalSelfAttention():

    test_tensor1 = torch.randn(2, 64, 1024)
    test_tensor2 = torch.randn(2, 256, 1024)

    attention = ATTENTION_REGISTRY.get("GatedPositionalSelfAttention")(dim=1024)
    out = attention(test_tensor1)
    assert out.shape == (2, 64, 1024)
    del attention

    attention = ATTENTION_REGISTRY.get("GatedPositionalSelfAttention")(
        dim=1024, num_heads=16
    )
    out = attention(test_tensor2)
    assert out.shape == (2, 256, 1024)
    del attention


def test_ConvVTAttention():
    test_tensor1 = torch.randn(16, 196, 384)
    attention = ATTENTION_REGISTRY.get("ConvVTAttention")(384, 128, 4, 14)
    out = attention(test_tensor1)
    assert out.shape == torch.Size([16, 196, 128])
    del attention

    test_tensor1 = torch.randn(16, 196, 384)
    attention = ATTENTION_REGISTRY.get("ConvVTAttention")(384, 128, 4, 14, method="avg")
    out = attention(test_tensor1)
    assert out.shape == torch.Size([16, 196, 128])
    del attention



import torch
import torch.nn as nn
import torch.nn.functional as F

from ...common import BaseClassificationModel
from ...decoder import MLPDecoder
from ...encoder import CVTEmbedding, PosEmbedding, VanillaEncoder
from ...utils import MODEL_REGISTRY, pair


@MODEL_REGISTRY.register()
class CCT(BaseClassificationModel):
    """
    Implementation of Escaping the Big Data Paradigm with Compact Transformers:
    https://arxiv.org/abs/2104.05704

    Parameters:
    ------------
    img_size: int
        Size of the image
    patch_size: int
        Size of the single patch in the image
    in_channels: int
        Number of input channels in image
    seq_pool:bool
        Whether to use sequence pooling or not
    embedding_dim: int
        Patch embedding dimension
    num_layers: int
        Number of Encoders in encoder block
    num_heads: int
        Number of heads in each transformer layer
    mlp_ratio:float
        Ratio of mlp heads to embedding dimension
    n_classes: int
        Number of classes for classification
    p_dropout: float
        Dropout probability
    attn_dropout: float
        Dropout probability
    drop_path: float
        Stochastic depth rate, default is 0.1
    positional_embedding: str
        One of the string values {'learnable','sine','None'}, default is learnable
    decoder_config: tuple(int) or int
        Configuration of the decoder. If None, the default configuration is used.
    pooling_kernel_size: int or tuple(int)
        Size of the kernel in MaxPooling operation
    pooling_stride: int or tuple(int)
        Stride of MaxPooling operation
    pooling_padding: int
        Padding in MaxPooling operation
    """

    def __init__(
        self,
        img_size=224,
        patch_size=4,
        in_channels=3,
        seq_pool=True,
        embedding_dim=768,
        num_layers=1,
        head_dim=96,
        num_heads=1,
        mlp_ratio=4.0,
        n_classes=1000,
        p_dropout=0.1,
        attn_dropout=0.1,
        drop_path=0.1,
        positional_embedding="learnable",
        decoder_config=(
            768,
            1024,
        ),
        pooling_kernel_size=3,
        pooling_stride=2,
        pooling_padding=1,
    ):
        super().__init__(
            img_size=img_size,
            patch_size=patch_size,
        )

        assert (
            img_size % patch_size == 0
        ), f"Image size ({img_size}) has to be divisible by patch size ({patch_size})"

        img_size = pair(img_size)
        self.in_channels = in_channels

        self.embedding = CVTEmbedding(
            in_channels=in_channels,
            out_channels=embedding_dim,
            kernel_size=patch_size,
            stride=patch_size,
            padding=0,
            max_pool=True,
            pooling_kernel_size=pooling_kernel_size,
            pooling_stride=pooling_stride,
            pooling_padding=pooling_padding,
            activation=nn.ReLU,
            num_conv_layers=1,
            conv_bias=True,
        )

        positional_embedding = (
            positional_embedding
            if positional_embedding in ["sine", "learnable", "none"]
            else "sine"
        )
        hidden_dim = int(embedding_dim * mlp_ratio)
        self.embedding_dim = embedding_dim
        self.sequence_length = self.embedding.sequence_length(
            n_channels=in_channels, height=img_size[0], width=img_size[1]
        )
        self.seq_pool = seq_pool

        assert (
            self.sequence_length is not None or positional_embedding == "none"
        ), f"Positional embedding is set to {positional_embedding} and the sequence length was not specified."

        if not seq_pool:
            self.sequence_length += 1
            self.class_emb = nn.Parameter(
                torch.zeros(1, 1, self.embedding_dim), requires_grad=True
            )
        else:
            self.attention_pool = nn.Linear(self.embedding_dim, 1)

        if positional_embedding != "none":
            self.positional_emb = PosEmbedding(
                self.sequence_length,
                dim=embedding_dim,
                drop=p_dropout,
                sinusoidal=True if positional_embedding is "sine" else False,
            )
        else:
            self.positional_emb = None

        dpr = [x.item() for x in torch.linspace(0, drop_path, num_layers)]
        self.encoder_blocks = nn.ModuleList(
            [
                VanillaEncoder(
                    embedding_dim=embedding_dim,
                    num_heads=num_heads,
                    depth=1,
                    head_dim=head_dim,
                    mlp_dim=hidden_dim,
                    p_dropout=p_dropout,
                    attn_dropout=attn_dropout,
                    drop_path_rate=dpr[i],
                )
                for i in range(num_layers)
            ]
        )
        if decoder_config is not None:

            if not isinstance(decoder_config, list) and not isinstance(
                decoder_config, tuple
            ):
                decoder_config = [decoder_config]

            assert (
                decoder_config[0] == embedding_dim
            ), f"Configurations do not match for MLPDecoder, First element of `decoder_config` expected to be {embedding_dim}, got {decoder_config[0]} "
            self.decoder = MLPDecoder(config=decoder_config, n_classes=n_classes)

        else:
            self.decoder = MLPDecoder(config=embedding_dim, n_classes=n_classes)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of size `n_classes`

        """
        x = self.embedding(x)

        if self.positional_emb is None and x.size(1) < self.sequence_length:
            x = F.pad(
                x, (0, 0, 0, self.in_channels - x.size(1)), mode="constant", value=0
            )

        if not self.seq_pool:
            cls_token = self.class_emb.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_token, x), dim=1)

        if self.positional_emb is not None:
            x = self.positional_emb(x)

        for blk in self.encoder_blocks:
            x = blk(x)

        if self.seq_pool:
            x = torch.matmul(
                F.softmax(self.attention_pool(x), dim=1).transpose(-1, -2), x
            ).squeeze(-2)
        else:
            x = x[:, 0]

        x = self.decoder(x)

        return x



import torch.nn as nn
import torch.utils.checkpoint as checkpoint
from timm.models.layers import DropPath

from ..attention.window import WindowAttention
from ..utils import (
    ENCODER_REGISTRY,
    create_mask,
    cyclicshift,
    pair,
    window_partition,
    window_reverse,
)
from .nn import FeedForward


@ENCODER_REGISTRY.register()
class SwinEncoderBlock(nn.Module):
    """

    Parameters
    ----------
    dim: int
        Number of the input channels
    input_resolution: int or tuple[int]
        Input resolution of patches
    num_heads: int
        Number of attention heads
    window_size: int
        Window size
    shift_size: int
        Shift size for Shifted Window Masked Self Attention (SW_MSA)
    mlp_ratio: float
        Ratio of MLP hidden dimension to embedding dimension
    qkv_bias: bool, default= True
        Whether to add a bias vector to the q,k, and v matrices
    qk_scale: float, Optional

    p_dropout: float
        Dropout rate
    attn_dropout: float
        Dropout rate
    drop_path_rate: float
        Stochastic depth rate
    norm_layer:nn.Module
        Normalization layer, default is `nn.LayerNorm`

    """

    def __init__(
        self,
        dim,
        input_resolution,
        num_heads,
        window_size=7,
        shift_size=0,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
    ):
        super(SwinEncoderBlock, self).__init__()

        self.dim = dim
        self.input_resolution = pair(input_resolution)
        self.num_heads = num_heads
        self.window_size = window_size
        self.mlp_ratio = mlp_ratio
        self.shift_size = shift_size
        hidden_dim = int(dim * mlp_ratio)

        if min(self.input_resolution) <= self.window_size:
            self.shift_size = 0
            self.window_size = min(self.input_resolution)

        assert (
            0 <= self.shift_size < window_size
        ), "shift size must range from 0 to window size"

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim=dim,
            window_size=pair(self.window_size),
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_dropout=attn_dropout,
            proj_dropout=p_dropout,
        )

        self.drop_path = (
            DropPath(drop_path_rate) if drop_path_rate > 0.0 else nn.Identity()
        )
        self.norm2 = norm_layer(dim)
        self.mlp = FeedForward(dim=dim, hidden_dim=hidden_dim, p_dropout=p_dropout)

        if self.shift_size > 0:
            attn_mask = create_mask(
                self.window_size,
                self.shift_size,
                H=self.input_resolution[0],
                W=self.input_resolution[1],
            )
        else:
            attn_mask = None

        self.register_buffer("attn_mask", attn_mask)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor

        Returns
        ----------
        torch.Tensor
            Returns output tensor

        """

        H, W = self.input_resolution
        B, L, C = x.shape
        assert L == H * W, "Input tensor shape not compatible"

        skip_connection = x

        x = self.norm1(x)
        x = x.view(B, H, W, C)

        if self.shift_size > 0:
            shifted_x = cyclicshift(x, shift_size=-self.shift_size)
        else:
            shifted_x = x

        x_windows = window_partition(shifted_x, self.window_size).view(
            -1, self.window_size * self.window_size, C
        )

        attn_windows = self.attn(x_windows, mask=self.attn_mask).view(
            -1, self.window_size, self.window_size, C
        )

        shifted_x = window_reverse(attn_windows, self.window_size, H, W)

        if self.shift_size > 0:
            x = cyclicshift(shifted_x, shift_size=self.shift_size).view(B, H * W, C)
        else:
            x = shifted_x.view(B, H * W, C)

        x = skip_connection + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


@ENCODER_REGISTRY.register()
class SwinEncoder(nn.Module):
    """
    dim: int
        Number of input channels.
    input_resolution: tuple[int]
        Input resolution.
    depth: int
        Number of blocks.
    num_heads: int
        Number of attention heads.
    window_size: int
        Local window size.
    mlp_ratio: float
        Ratio of MLP hidden dim to embedding dim.
    qkv_bias: bool, default is True
       Whether to add a bias vector to the q,k, and v matrices
    qk_scale: float, optional
        Override default qk scale of head_dim ** -0.5 in Window Attention if set
    p_dropout: float,
        Dropout rate.
    attn_dropout: float, optional
        Attention dropout rate
    drop_path_rate: float or tuple[float]
        Stochastic depth rate.
    norm_layer: nn.Module
        Normalization layer. default is nn.LayerNorm
    downsample: nn.Module, optional
        Downsample layer(like PatchMerging) at the end of the layer, default is None

    """

    def __init__(
        self,
        dim,
        input_resolution,
        depth,
        num_heads,
        window_size,
        mlp_ratio=4.0,
        qkv_bias=True,
        qkv_scale=None,
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path=0.0,
        norm_layer=nn.LayerNorm,
        downsample=None,
        use_checkpoint=False,
    ):
        super(SwinEncoder, self).__init__()

        self.dim = dim
        self.input_resolution = input_resolution
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        self.blocks = nn.ModuleList(
            [
                SwinEncoderBlock(
                    dim=dim,
                    input_resolution=input_resolution,
                    num_heads=num_heads,
                    window_size=window_size,
                    shift_size=0 if (i % 2 == 0) else window_size // 2,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    qk_scale=qkv_scale,
                    p_dropout=p_dropout,
                    attn_dropout=attn_dropout,
                    drop_path_rate=drop_path[i]
                    if isinstance(drop_path, list)
                    else drop_path,
                    norm_layer=norm_layer,
                )
                for i in range(depth)
            ]
        )
        if downsample is not None:
            self.downsample = downsample(
                input_resolution, dim=dim, norm_layer=norm_layer
            )
        else:
            self.downsample = None

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor

        Returns
        ----------
        torch.Tensor
            Returns output tensor

        """

        for blk in self.blocks:
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x)
            else:
                x = blk(x)

        if self.downsample is not None:
            x = self.downsample(x)

        return x



import torch.nn as nn
from timm.models.layers import DropPath

from ..attention import GatedPositionalSelfAttention
from ..functional import PreNorm
from ..utils import ENCODER_REGISTRY
from .nn import FeedForward
from .vanilla import VanillaEncoder


@ENCODER_REGISTRY.register()
class ConViTEncoder(VanillaEncoder):
    """

    Parameters
    ----------
    embedding_dim: int
        Dimension of the embedding
    depth: int
        Number of self-attention layers
    num_heads: int
        Number of the attention heads
    head_dim: int
        Dimension of each head
    mlp_dim: int
        Dimension of the hidden layer in the feed-forward layer
    p_dropout: float
        Dropout Probability
    attn_dropout: float
        Dropout Probability
    drop_path_rate: float
        Stochastic drop path rate
    """

    def __init__(
        self,
        embedding_dim,
        depth,
        num_heads,
        head_dim,
        mlp_dim,
        p_dropout=0,
        attn_dropout=0,
        drop_path_rate=0,
    ):
        super().__init__(
            embedding_dim,
            depth,
            num_heads,
            head_dim,
            mlp_dim,
            p_dropout,
            attn_dropout,
            drop_path_rate,
        )
        self.encoder = nn.ModuleList([])

        for _ in range(depth):
            self.encoder.append(
                nn.ModuleList(
                    [
                        PreNorm(
                            dim=embedding_dim,
                            fn=GatedPositionalSelfAttention(
                                dim=embedding_dim,
                                num_heads=num_heads,
                                head_dim=head_dim,
                                p_dropout=attn_dropout,
                            ),
                        ),
                        PreNorm(
                            dim=embedding_dim,
                            fn=FeedForward(
                                dim=embedding_dim,
                                hidden_dim=mlp_dim,
                                p_dropout=p_dropout,
                            ),
                        ),
                    ]
                )
            )
        self.drop_path = (
            DropPath(drop_prob=drop_path_rate)
            if drop_path_rate > 0.0
            else nn.Identity()
        )



import torch

from vformer.models import VanillaViT
from vformer.viz import ViTAttentionGradRollout, ViTAttentionRollout


def test_attention_rollout():
    img = torch.randn(1, 3, 256, 256)
    layer = "attend"
    model = VanillaViT(img_size=256, patch_size=32, n_classes=10, in_channels=3)
    model_attention_rollout_mean = ViTAttentionRollout(model, layer)
    model_attention_rollout_max = ViTAttentionRollout(model, layer, "max")
    model_attention_rollout_min = ViTAttentionRollout(model, layer, "min")
    _ = model(img)
    _ = model_attention_rollout_mean(img)
    _ = model_attention_rollout_max(img)
    _ = model_attention_rollout_min(img)


def test_attention_grad_rollout():
    img = torch.randn(1, 3, 256, 256)
    layer = "attend"
    model = VanillaViT(img_size=256, patch_size=32, n_classes=10, in_channels=3)
    model_attention_grad_rollout = ViTAttentionGradRollout(model, layer)
    _ = model_attention_grad_rollout(img, 0)



"""
Adapted from Detectron2 (https://github.com/facebookresearch/detectron2)
"""


class Registry:
    """
    Class to register objects and then retrieve them by name.
    Parameters
    ----------
    name : str
        Name of the registry
    """

    def __init__(self, name):

        self._name = name
        self._obj_map = {}

    def _do_register(self, name, obj):

        assert (
            name not in self._obj_map
        ), f"An object named '{name}' was already registered in '{self._name}' registry!"

        self._obj_map[name] = obj

    def register(self, obj=None, name=None):
        """
        Method to register an object in the registry
        Parameters
        ----------
        obj : object, optional
            Object to register, defaults to None (which will return the decorator)
        name : str, optional
            Name of the object to register, defaults to None (which will use the name of the object)
        """

        if obj is None:

            def deco(func_or_class, name=name):
                if name is None:
                    name = func_or_class.__name__
                self._do_register(name, func_or_class)
                return func_or_class

            return deco

        if name is None:  # pragma: no cover
            name = obj.__name__

        self._do_register(name, obj)  # pragma: no cover

    def get(self, name):
        """
        Method to retrieve an object from the registry
        Parameters
        ----------
        name : str
            Name of the object to retrieve
        Returns
        -------
        object
            Object registered under the given name
        """

        ret = self._obj_map.get(name)
        if ret is None:  # pragma: no cover
            raise KeyError(
                f"No object named '{name}' found in '{self._name}' registry!"
            )

        return ret

    def get_list(self):
        """
        Method to retrieve all objects from the registry
        Returns
        -------
        list
            List of all objects registered in the registry
        """

        return list(self._obj_map.keys())

    def __contains__(self, name):
        return name in self._obj_map  # pragma: no cover

    def __iter__(self):
        return iter(self._obj_map.items())  # pragma: no cover


ATTENTION_REGISTRY = Registry("ATTENTION")
DECODER_REGISTRY = Registry("DECODER")
ENCODER_REGISTRY = Registry("ENCODER")
MODEL_REGISTRY = Registry("MODEL")



# thanks to https://github.com/isl-org/DPT

import math

import torch
import torch.nn.functional as F


def forward_flex(self, x):
    b, c, h, w = x.shape

    pos_embed = self.model._resize_pos_embed(
        self.model.pos_embedding.pos_embed,
        h // self.model.patch_size[1],
        w // self.model.patch_size[0],
    )

    B = x.shape[0]
    x = self.model.patch_embedding.patch_embedding(x)
    cls_tokens = self.model.cls_token.expand(B, -1, -1)
    x = torch.cat((cls_tokens, x), dim=1)
    x = x + pos_embed
    x = self.model.pos_embedding.pos_drop(x)
    x = self.model.encoder(x)
    return x


def _resize_pos_embed(self, posemb, gs_h, gs_w):
    posemb_tok, posemb_grid = (
        posemb[:, : self.start_index],
        posemb[0, self.start_index :],
    )

    gs_old = int(math.sqrt(len(posemb_grid)))

    posemb_grid = posemb_grid.reshape(1, gs_old, gs_old, -1).permute(0, 3, 1, 2)
    posemb_grid = F.interpolate(posemb_grid, size=(gs_h, gs_w), mode="bilinear")
    posemb_grid = posemb_grid.permute(0, 2, 3, 1).reshape(1, gs_h * gs_w, -1)

    posemb = torch.cat([posemb_tok, posemb_grid], dim=1)

    return posemb



import torch
import torch.nn as nn


class CVTEmbedding(nn.Module):
    """
    This class converts the image patches to tensors. Size of the image patches is controlled by `stride` parameter.

    Parameters
    ----------
    kernel_size: int or tuple
        Size of the kernel used in convolution
    stride: int or tuple
        Stride of the convolution operation
    padding: int
        Padding to all sides of the input
    pooling_kernel_size: int or tuple(int)
        Size of the kernel used in  MaxPool2D,default is 3
    pooling_stride: int or tuple(int)
        Size of the stride in MaxPool2D, default is 2
    pooling_padding: int
        padding in the MaxPool2D
    num_conv_layers: int
        Number of Convolution layers in the encoder,default is 1
    in_channels: int
        Number of input channels in image, default is 3
    out_channels: int
        Number of output channels, default is 64
    in_planes: int
        This will be number of channels in the self.conv_layer's convolution except 1st layer and last layer.
    activation: nn.Module, optional
        Activation Layer, default is None
    max_pool: bool
        Whether to have max-pooling or not, change this parameter to False when using in CVT model, default is True
    conv_bias:bool
        Whether to add learnable bias in the convolution operation,default is False
    """

    def __init__(
        self,
        kernel_size,
        stride,
        padding,
        pooling_kernel_size=3,
        pooling_stride=2,
        pooling_padding=1,
        num_conv_layers=1,
        in_channels=3,
        out_channels=64,
        in_planes=64,
        activation=None,
        max_pool=True,
        conv_bias=False,
    ):
        super(CVTEmbedding, self).__init__()

        n_filter_list = (
            [in_channels]
            + [in_planes for _ in range(num_conv_layers - 1)]
            + [out_channels]
        )
        self.conv_layers = nn.ModuleList([])
        for i in range(num_conv_layers):
            self.conv_layers.append(
                nn.ModuleList(
                    [
                        nn.Conv2d(
                            n_filter_list[i],
                            n_filter_list[i + 1],
                            kernel_size=kernel_size,
                            stride=stride,
                            padding=padding,
                            bias=conv_bias,
                        ),
                        nn.Identity() if activation is None else activation(),
                        nn.MaxPool2d(
                            kernel_size=pooling_kernel_size,
                            stride=pooling_stride,
                            padding=pooling_padding,
                        )
                        if max_pool
                        else nn.Identity(),
                    ]
                )
            )

        self.flatten = nn.Flatten(2, 3)

    def sequence_length(self, n_channels=3, height=224, width=224):
        return self.forward(torch.zeros((1, n_channels, height, width))).shape[1]

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.tensor
            Input tensor

        Returns
        -----------
        torch.Tensor
            Returns output tensor (embedding) by applying multiple convolution and max-pooling operations on input tensor

        """

        for conv2d, activation, maxpool in self.conv_layers:
            x = maxpool(activation(conv2d(x)))

            return self.flatten(x).transpose(-2, -1)



import torch
import torch.nn as nn

from ..attention import VanillaSelfAttention
from ..functional.norm import PreNorm
from ..utils.registry import ENCODER_REGISTRY
from .nn import FeedForward


@ENCODER_REGISTRY.register()
class ViViTEncoderBlock(nn.Module):
    """For model 3 only """

    def __init__(
        self, dim, num_heads, head_dim, p_dropout, out_dim=None, hidden_dim=None
    ):
        super(ViViTEncoderBlock, self).__init__()

        self.temporal_attention = PreNorm(
            dim=dim, fn=VanillaSelfAttention(dim, num_heads, head_dim, p_dropout)
        )
        self.spatial_attention = PreNorm(
            dim=dim, fn=VanillaSelfAttention(dim, num_heads, head_dim, p_dropout)
        )

        self.mlp = FeedForward(dim=dim, hidden_dim=hidden_dim, out_dim=out_dim)

    def forward(self, x):

        b, n, s, d = x.shape
        x = torch.flatten(x, start_dim=0, end_dim=1)  # 1×nt·nh·nw·d --> nt×nh·nw·d

        x = self.spatial_attention(x) + x

        x = x.reshape(b, n, s, d).transpose(1, 2)
        x = torch.flatten(x, start_dim=0, end_dim=1)  # nt×nh·nw·d --> nh·nw×nt·d

        x = self.temporal_attention(x) + x

        x = self.mlp(x) + x

        x = x.reshape(
            b, n, s, d
        )  # reshaping because this block is used for several depths in ViViTEncoder class and Next layer will expect the x in proper shape

        return x


@ENCODER_REGISTRY.register()
class ViViTEncoder(nn.Module):
    """model 3 only"""

    def __init__(
        self, dim, num_heads, head_dim, p_dropout, depth, out_dim=None, hidden_dim=None
    ):
        super(ViViTEncoder, self).__init__()
        self.encoder = nn.ModuleList()

        for _ in range(depth):
            self.encoder.append(
                ViViTEncoderBlock(
                    dim, num_heads, head_dim, p_dropout, out_dim, hidden_dim
                )
            )

    def forward(self, x):

        b = x.shape[0]

        for blk in self.encoder:
            x = blk(x)

        x = x.reshape(b, -1, x.shape[-1])

        return x



import torch
import torch.nn as nn

from ..utils import pair


class PatchMerging(nn.Module):
    """

    Parameters
    ----------
    input_resolution: int or tuple[int]
        Resolution of input features
    dim : int
    """

    def __init__(self, input_resolution, dim, norm_layer=nn.LayerNorm):
        super(PatchMerging, self).__init__()
        self.input_resolution = pair(input_resolution)
        self.dim = dim
        self.reduction = nn.Linear(4 * self.dim, 2 * self.dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x):
        H, W = self.input_resolution
        B, L, C = x.shape
        assert L == H * W, "input feature has wrong size"
        assert H % 2 == 0 and W % 2 == 0, f"x size ({H}*{W}) are not even."

        x = x.view(B, H, W, C)

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)

        return x



import torch
import torch.nn as nn
from einops import rearrange
from timm.models.layers import DropPath, trunc_normal_

from ..attention.convvt import ConvVTAttention
from ..encoder.nn import FeedForward
from ..utils import ENCODER_REGISTRY
from .embedding.convvt import ConvEmbedding


@ENCODER_REGISTRY.register()
class ConvVTStage(nn.Module):
    """
    Implementation of a Stage in CVT

    Parameters
    ----------
    patch_size: int
        Size of patch, default is 16
    patch_stride: int
        Stride of patch, default is 4
    patch_padding: int
        Padding for patch, default is 0
    in_channels:int
        Number of input channels in image, default is 3
    img_size: int
        Size of the image, default is 224
    embedding_dim: int
        Embedding dimensions, default is 64
    depth: int
        Number of CVT Attention blocks in each stage, default is 1
    num_heads: int
        Number of heads in attention, default is 6
    mlp_ratio: float
        Feature dimension expansion ratio in MLP, default is 4.0
    p_dropout: float
        Probability of dropout in MLP, default is 0.0
    attn_dropout: float
        Probability of dropout in attention, default is 0.0
    drop_path_rate: float
        Probability for droppath, default is 0.0
    with_cls_token: bool
        Whether to include classification token, default is False
    kernel_size: int
        Size of kernel, default is 3
    padding_q: int
        Size of padding in q, default is 1
    padding_kv: int
        Size of padding in kv, default is 2
    stride_kv: int
        Stride in kv, default is 2
    stride_q: int
        Stride in q, default is 1
    init: str ('trunc_norm' or 'xavier')
        Initialization method, default is 'trunc_norm'
    """

    def __init__(
        self,
        patch_size=7,
        patch_stride=4,
        patch_padding=0,
        in_channels=3,
        embedding_dim=64,
        depth=1,
        p_dropout=0.0,
        drop_path_rate=0.0,
        with_cls_token=False,
        init="trunc_norm",
        **kwargs
    ):
        super().__init__()

        self.patch_embed = ConvEmbedding(
            patch_size=patch_size,
            in_channels=in_channels,
            embedding_dim=embedding_dim,
            stride=patch_stride,
            padding=patch_padding,
        )

        if with_cls_token:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embedding_dim))
        else:
            self.cls_token = None

        self.pos_drop = nn.Dropout(p=p_dropout)
        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, depth)
        ]  # stochastic depth decay rule

        blocks = []
        for j in range(depth):
            blocks.append(
                ConvVTBlock(
                    dim_in=embedding_dim,
                    dim_out=embedding_dim,
                    p_dropout=p_dropout,
                    with_cls_token=with_cls_token,
                    drop_path=dpr[j],
                    **kwargs
                )
            )
        self.blocks = nn.ModuleList(blocks)

        if self.cls_token is not None:
            trunc_normal_(self.cls_token, std=0.02)

        if init == "xavier":
            self.apply(self._init_weights_xavier)
        elif init == "trunc_norm":
            self.apply(self._init_weights_trunc_normal)
        else:
            raise ValueError("Init method {} not found".format(init))

    def _init_weights_trunc_normal(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm2d)):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def _init_weights_xavier(self, m):
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm2d)):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, x):

        x = self.patch_embed(x)
        B, C, H, W = x.size()

        x = rearrange(x, "b c h w -> b (h w) c")

        cls_tokens = None
        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(B, -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)

        x = self.pos_drop(x)

        for blk in self.blocks:
            x = blk(x)

        if self.cls_token is not None:
            cls_tokens, x = torch.split(x, [1, H * W], 1)
        x = rearrange(x, "b (h w) c -> b c h w", h=H, w=W)

        return x, cls_tokens


class ConvVTBlock(nn.Module):
    """
    Implementation of a Attention MLP block in CVT

    Parameters:
    ------------
    dim_in: int
        Input dimensions
    dim_out: int
        Output dimensions
    num_heads: int
        Number of heads in attention
    img_size: int
        Size of image
    mlp_ratio: float
        Feature dimension expansion ratio in MLP, default is 4.
    p_dropout: float
        Probability of dropout in MLP, default is 0.0
    attn_dropout: float
        Probability of dropout in attention, default is 0.0
    drop_path: float
        Probability of droppath, default is 0.0
    with_cls_token: bool
        Whether to include classification token, default is False
    """

    def __init__(
        self, dim_in, dim_out, mlp_ratio=4.0, p_dropout=0.0, drop_path=0.0, **kwargs
    ):
        super().__init__()

        self.norm1 = nn.LayerNorm(dim_in)
        self.attn = ConvVTAttention(dim_in, dim_out, **kwargs)

        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = nn.LayerNorm(dim_out)

        dim_mlp_hidden = int(dim_out * mlp_ratio)
        self.mlp = FeedForward(
            dim=dim_out, hidden_dim=dim_mlp_hidden, p_dropout=p_dropout
        )

    def forward(self, x):

        res = x

        x = self.norm1(x)
        attn = self.attn(x)
        x = res + self.drop_path(attn)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x



import torch.nn as nn

from ..utils import DECODER_REGISTRY


@DECODER_REGISTRY.register()
class MLPDecoder(nn.Module):
    """
    Parameters
    ----------
    config : int or tuple or list
        Configuration of the hidden layer(s)
    n_classes : int
        Number of classes for classification
    """

    def __init__(self, config=(1024,), n_classes=10):
        super(MLPDecoder, self).__init__()

        self.decoder = nn.ModuleList()

        if not isinstance(config, list) and not isinstance(config, tuple):
            config = [config]

        if len(config) > 1:
            for i in range(len(config) - 1):

                self.decoder.append(nn.LayerNorm(config[i]))
                self.decoder.append(nn.Linear(config[i], config[i + 1]))

        self.decoder.append(nn.LayerNorm(config[-1]))
        self.decoder.append(nn.Linear(config[-1], n_classes))

        self.decoder = nn.Sequential(*self.decoder)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns output tensor of size `n_classes`, Note that `torch.nn.Softmax` is not applied to the output tensor.

        """
        return self.decoder(x)



import torch
import torch.nn as nn
from timm.models.layers import trunc_normal_

from ..utils import ATTENTION_REGISTRY, get_relative_position_bias_index, pair


@ATTENTION_REGISTRY.register()
class WindowAttention(nn.Module):
    """
    Parameters
    ----------
    dim: int
        Number of input channels.
    window_size : int or tuple[int]
        The height and width of the window.
    num_heads: int
        Number of attention heads.
    qkv_bias :bool, default is True
        If True, add a learnable bias to query, key, value.
    qk_scale: float, optional
        Override default qk scale of head_dim ** -0.5 if set
    attn_dropout: float, optional
        Dropout rate
    proj_dropout: float, optional
        Dropout rate

    """

    def __init__(
        self,
        dim,
        window_size,
        num_heads,
        qkv_bias=True,
        qk_scale=None,
        attn_dropout=0.0,
        proj_dropout=0.0,
    ):
        super(WindowAttention, self).__init__()

        self.dim = dim
        self.window_size = pair(window_size)
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = qk_scale or self.head_dim ** -0.5
        self.qkv_bias = True
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros(
                (2 * self.window_size[0] - 1) * (2 * self.window_size[1] - 1), num_heads
            )
        )
        relative_position_index = get_relative_position_bias_index(self.window_size)
        self.register_buffer("relative_position_index", relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.to_out_1 = nn.Sequential(nn.Softmax(dim=-1), nn.Dropout(attn_dropout))
        self.to_out_2 = nn.Sequential(nn.Linear(dim, dim), nn.Dropout(proj_dropout))
        trunc_normal_(self.relative_position_bias_table, std=0.2)

    def forward(self, x, mask=None):
        """

        Parameters
        ----------
        x: torch.Tensor
            input Tensor
        mask: torch.Tensor
            Attention mask used for shifted window attention, if None, window attention will be used,
            else attention mask will be taken into consideration.
            for better understanding you may refer `this <https://github.com/microsoft/Swin-Transformer/issues/38>`

        Returns
        ----------
        torch.Tensor
            Returns output tensor by applying Window-Attention or Shifted-Window-Attention on input tensor

        """

        B_, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(
                B_,
                N,
                3,
                self.num_heads,
                C // self.num_heads,
            )
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = qkv[0], qkv[1], qkv[2]
        q = q * self.scale
        attn = q @ k.transpose(-2, -1)

        relative_position_bias = (
            self.relative_position_bias_table[self.relative_position_index.view(-1)]
            .view(
                self.window_size[0] * self.window_size[1],
                self.window_size[0] * self.window_size[1],
                -1,
            )
            .permute(2, 0, 1)
            .contiguous()
        )
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(
                1
            ).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)

        attn = self.to_out_1(attn)
        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.to_out_2(x)

        return x



import torch
import torch.nn as nn
from einops import rearrange


class ConvEmbedding(nn.Module):
    """
    This class converts images to tensors.

    Parameters
    ----------
    patch_size: int, default is 7
        Size of a patch
    in_channels: int, default is 3
        Number of input channels
    embedding_dim: int, default is 64
        Dimension of hidden layer
    stride: int or tuple, default is 4
        Stride of the convolution operation
    padding: int, default is 2
        Padding to all sides of the input
    """

    def __init__(
        self, patch_size=7, in_channels=3, embedding_dim=64, stride=4, padding=2
    ):
        super().__init__()
        self.patch_size = (patch_size, patch_size)

        self.proj = nn.Conv2d(
            in_channels,
            embedding_dim,
            kernel_size=self.patch_size,
            stride=stride,
            padding=padding,
        )
        self.norm = nn.LayerNorm(embedding_dim)

    def forward(self, x):
        """
        Parameters
        ----------
        x: torch.tensor
            Input tensor

        Returns
        -----------
        torch.Tensor
            Returns output tensor (embedding) by applying a convolution operations on input tensor
        """
        x = self.proj(x)
        B, C, H, W = x.shape
        x = rearrange(x, "b c h w -> b (h w) c")
        x = self.norm(x)
        x = rearrange(x, "b (h w) c -> b c h w", h=H, w=W)

        return x



import torch.nn as nn


class DWConv(nn.Module):
    """
    Depth Wise Convolution

    Parameters
    ----------
    dim: int
        Dimension of the input tensor
    kernel_size_dwconv: int,optional
        Size of the convolution kernel, default is 3
    stride_dwconv: int
        Stride of the convolution, default is 1
    padding_dwconv: int or tuple or str
        Padding added to all sides of the input, default is 1
    bias_dwconv:bool
        Whether to add learnable bias to the output,default is True.

    """

    def __init__(
        self,
        dim,
        kernel_size_dwconv=3,
        stride_dwconv=1,
        padding_dwconv=1,
        bias_dwconv=True,
    ):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(
            dim,
            dim,
            kernel_size=kernel_size_dwconv,
            stride=stride_dwconv,
            padding=padding_dwconv,
            bias=bias_dwconv,
            groups=dim,
        )

    def forward(self, x, H, W):
        """

        Parameters:
        ----------
        x: torch.Tensor
            Input tensor
        H: int
            Height of image patch
        W: int
            Width of image patch

        Returns:
        ----------
        torch.Tensor
            Returns output tensor after performing depth-wise convolution operation

        """

        B, N, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)

        x = x.flatten(2).transpose(1, 2)
        return x



import torch.nn as nn


class PreNorm(nn.Module):
    """
    Parameters
    ----------
    dim: int
        Dimension of the embedding
    fn:nn.Module
        Attention class
    """

    def __init__(self, dim, fn):
        super().__init__()

        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)



import torch
import torch.nn as nn

from ..attention import CrossAttention
from ..utils import ENCODER_REGISTRY
from .vanilla import VanillaEncoder


@ENCODER_REGISTRY.register()
class CrossEncoder(nn.Module):
    """

    Parameters
    ----------
    embedding_dim_s : int
        Dimension of the embedding of smaller patches, default is 1024
    embedding_dim_l : int
        Dimension of the embedding of larger patches, default is 1024
    attn_heads_s : int
        Number of self-attention heads for the smaller patches, default is 16
    attn_heads_l : int
        Number of self-attention heads for the larger patches, default is 16
    cross_head_s : int
        Number of cross-attention heads for the smaller patches, default is 8
    cross_head_l : int
        Number of cross-attention heads for the larger patches, default is 8
    head_dim_s : int
        Dimension of the head of the attention for the smaller patches, default is 64
    head_dim_l : int
        Dimension of the head of the attention for the larger patches, default is 64
    cross_dim_head_s : int
        Dimension of the head of the cross-attention for the smaller patches, default is 64
    cross_dim_head_l : int
        Dimension of the head of the cross-attention for the larger patches, default is 64
    depth_s : int
        Number of self-attention layers in encoder for the smaller patches, default is 6
    depth_l : int
        Number of self-attention layers in encoder for the larger patches, default is 6
    mlp_dim_s : int
        Dimension of the hidden layer in the feed-forward layer for the smaller patches, default is 2048
    mlp_dim_l : int
        Dimension of the hidden layer in the feed-forward layer for the larger patches, default is 2048
    p_dropout_s : float
        Dropout probability for the smaller patches, default is 0.0
    p_dropout_l : float
        Dropout probability for the larger patches, default is 0.0
    """

    def __init__(
        self,
        embedding_dim_s=1024,
        embedding_dim_l=1024,
        attn_heads_s=16,
        attn_heads_l=16,
        cross_head_s=8,
        cross_head_l=8,
        head_dim_s=64,
        head_dim_l=64,
        cross_dim_head_s=64,
        cross_dim_head_l=64,
        depth_s=6,
        depth_l=6,
        mlp_dim_s=2048,
        mlp_dim_l=2048,
        p_dropout_s=0.0,
        p_dropout_l=0.0,
    ):
        super().__init__()

        self.s = VanillaEncoder(
            embedding_dim_s,
            depth_s,
            attn_heads_s,
            head_dim_s,
            mlp_dim_s,
            p_dropout_s,
        )
        self.l = VanillaEncoder(
            embedding_dim_l,
            depth_l,
            attn_heads_l,
            head_dim_l,
            mlp_dim_l,
            p_dropout_l,
        )
        self.attend_s = CrossAttention(
            embedding_dim_s, embedding_dim_l, cross_head_s, cross_dim_head_s
        )
        self.attend_l = CrossAttention(
            embedding_dim_l, embedding_dim_s, cross_head_l, cross_dim_head_l
        )

    def forward(self, emb_s, emb_l):

        emb_s = self.s(emb_s)
        emb_l = self.l(emb_l)
        s_cls, s_patches = (lambda t: (t[:, 0:1, :], t[:, 1:, :]))(emb_s)
        l_cls, l_patches = (lambda t: (t[:, 0:1, :], t[:, 1:, :]))(emb_l)
        s_cls = self.attend_s(s_cls, l_patches)
        l_cls = self.attend_l(l_cls, s_patches)
        emb_l = torch.cat([l_cls, l_patches], dim=1)
        emb_s = torch.cat([s_cls, s_patches], dim=1)

        return emb_s, emb_l



import torch

from .utils import pair


def cyclicshift(input, shift_size, dims=None):
    """
    Parameters
    ----------
    input: torch.Tensor
        input tensor
    shift_size: int or tuple(int)
        Number of places by which input tensor is shifted
    dims: int or tuple(int),optional
        Axis along which to roll
    """

    return torch.roll(
        input, shifts=pair(shift_size), dims=(1, 2) if dims == None else dims
    )


def window_partition(x, window_size):
    """
    Parameters
    ----------
    x: torch.Tensor
        input tensor
    window_size: int
        window size
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
    windows = (
        x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    )

    return windows


def window_reverse(windows, window_size, H, W):
    """
    Parameters
    ----------
    windows: torch.Tensor
    window_size: int
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(
        B, H // window_size, W // window_size, window_size, window_size, -1
    )
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x


def get_relative_position_bias_index(window_size):
    """
    Parameters
    ----------
    window_size: int or tuple[int]
        Window size
    """
    window_size = pair(window_size)
    coords_h = torch.arange(window_size[0])
    coords_w = torch.arange(window_size[1])
    coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
    coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
    relative_coords = (
        coords_flatten[:, :, None] - coords_flatten[:, None, :]
    )  # 2, Wh*Ww, Wh*Ww
    relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
    relative_coords[:, :, 0] += window_size[0] - 1  # shift to start from 0
    relative_coords[:, :, 1] += window_size[1] - 1
    relative_coords[:, :, 0] *= 2 * window_size[1] - 1
    relative_position_index = relative_coords.sum(-1)
    return relative_position_index


def create_mask(window_size, shift_size, H, W):
    """
    Parameters
    ----------
    window_size: int
        Window Size
    shift_size: int
        Shift_size

    """
    img_mask = torch.zeros(1, H, W, 1)
    h_slices = (
        slice(0, -window_size),
        slice(-window_size, -shift_size),
        slice(-shift_size, None),
    )
    w_slices = (
        slice(0, -window_size),
        slice(-window_size, -shift_size),
        slice(-shift_size, None),
    )
    cnt = 0
    for h in h_slices:
        for w in w_slices:
            img_mask[:, h, w, :] = cnt
            cnt += 1
    mask_windows = window_partition(img_mask, window_size)
    mask_windows = mask_windows.view(-1, window_size * window_size)
    attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
    attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(
        attn_mask == 0, float(0.0)
    )
    return attn_mask



import torch

from vformer.utils import DECODER_REGISTRY

decoder_modules = DECODER_REGISTRY.get_list()


def test_MLPDecoder():

    test_tensor = torch.randn(2, 3, 100)

    decoder = DECODER_REGISTRY.get("MLPDecoder")(config=100, n_classes=10)
    out = decoder(test_tensor)
    assert out.shape == (2, 3, 10)
    del decoder

    decoder = DECODER_REGISTRY.get("MLPDecoder")(config=(100, 50), n_classes=10)
    out = decoder(test_tensor)
    assert out.shape == (2, 3, 10)
    del decoder

    decoder = DECODER_REGISTRY.get("MLPDecoder")(config=[100, 10], n_classes=5)
    out = decoder(test_tensor)
    assert out.shape == (2, 3, 5)


def test_SegmentationHead():

    test_tensor_segmentation_head_256 = [
        torch.randn([2, 64, 64, 64]),
        torch.randn([2, 128, 32, 32]),
        torch.randn([2, 256, 16, 16]),
        torch.randn([2, 512, 8, 8]),
    ]
    test_tensor_segmentation_head_224 = [
        torch.randn([2, 64, 56, 56]),
        torch.randn([2, 128, 28, 28]),
        torch.randn([2, 256, 14, 14]),
        torch.randn([2, 512, 7, 7]),
    ]
    test_tensor_segmentation_head = [
        torch.randn([3, 128, 96, 96]),
        torch.randn([3, 256, 48, 48]),
        torch.randn([3, 512, 24, 24]),
        torch.randn([3, 1024, 12, 12]),
    ]

    head = DECODER_REGISTRY.get("SegmentationHead")(
        out_channels=1,
    )
    out = head(test_tensor_segmentation_head_256)
    assert out.shape == (2, 1, 256, 256)

    head = DECODER_REGISTRY.get("SegmentationHead")(
        out_channels=10,
    )
    out = head(test_tensor_segmentation_head_224)
    assert out.shape == (2, 10, 224, 224)

    head = DECODER_REGISTRY.get("SegmentationHead")(
        out_channels=2, embed_dims=[128, 256, 512, 1024]
    )
    out = head(test_tensor_segmentation_head)
    assert out.shape == (3, 2, 384, 384)



import types

import torch
import torch.nn as nn

from ...utils.dpt_utils import _resize_pos_embed, forward_flex
from ...utils.registry import MODEL_REGISTRY

activations = {}


def get_activation(name):
    def hook(model, input, output):
        activations[name] = output

    return hook


attention = {}


def get_attention(name):
    def hook(module, input, output):
        x = input[0]
        B, N, C = x.shape
        qkv = (
            module.to_qkv(x)
            .reshape(B, N, 3, module.num_heads, C // module.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = (
            qkv[0],
            qkv[1],
            qkv[2],
        )  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * module.scale

        attn = attn.softmax(dim=-1)  # [:,:,1,1:]
        attention[name] = attn

    return hook


@MODEL_REGISTRY.register()
class DPTDepth(nn.Module):
    """
    Implementation of " Vision Transformers for Dense Prediction "
    https://arxiv.org/abs/2103.13413

    Parameters
    ----------
    backbone:str
        Name of ViT model to be used as backbone, must be one of {`vitb16`,`vitl16`,`vit_tiny`}
    in_channels: int
        Number of channels in input image, default is 3
    img_size: tuple[int]
        Input image size, default is (384,384)
    readout:str
        Method to handle the `readout_token` or `cls_token`
        Must be one of {`add`, `ignore`,`project`}, default is `project`
    hooks: list[int]
        List representing index of encoder blocks on which hooks will be registered.
        These hooks extract features from different ViT blocks, eg attention, default is (2,5,8,11).
    channels_last: bool
        Alters the memory format of storing tensors, default is False,
        For more information visit, this `blogpost<https://pytorch.org/tutorials/intermediate/memory_format_tutorial.html>`
    use_bn:bool
        If True, BatchNormalisation is used in `FeatureFusionBlock_custom`, default is False
    enable_attention_hooks:bool
        If True, `get_attention` hook is registered, default is false
    non_negative:bool
        If True, Relu operation will be applied in `DPTDepth.model.head` block, default is True
    invert:bool
        If True, forward pass output of `DPTDepth.model.head` will be transformed (inverted)
        according to `scale` and `shift` parameters, default is False
    scale:float
        Float value that will be multiplied with forward pass output from `DPTDepth.model.head`, default is 1.0
    shift:float
        Float value that will be added with forward pass output from `DPTDepth.model.head` after scaling, default is 0.0
    """

    def __init__(
        self,
        backbone,
        in_channels=3,
        img_size=(384, 384),
        readout="project",
        hooks=(2, 5, 8, 11),
        channels_last=False,
        use_bn=False,
        enable_attention_hooks=False,
        non_negative=True,
        scale=1.0,
        shift=0.0,
        invert=False,
    ):
        super(DPTDepth, self).__init__()

        self.channels_last = channels_last
        self.use_bn = use_bn
        self.enable_attention_hooks = enable_attention_hooks
        self.non_negative = non_negative
        self.scale = scale
        self.shift = shift
        self.invert = invert
        start_index = 1

        if backbone == "vitb16":

            scratch_in_features = (
                96,
                192,
                384,
                768,
            )
            self.model = MODEL_REGISTRY.get("VanillaViT")(
                img_size=img_size,
                patch_size=16,
                embedding_dim=768,
                head_dim=64,
                depth=12,
                attn_heads=12,
                encoder_mlp_dim=768,
                n_classes=10,
                in_channels=in_channels,
            )
            hooks = (2, 5, 8, 11) if hooks is None else hooks
            self.vit_features = 768

        elif backbone == "vitl16":

            scratch_in_features = (256, 512, 1024, 1024)
            self.model = MODEL_REGISTRY.get("VanillaViT")(
                img_size=img_size,
                patch_size=16,
                embedding_dim=1024,
                head_dim=64,
                depth=24,
                attn_heads=16,
                encoder_mlp_dim=1024,
                n_classes=10,
                in_channels=in_channels,
            )
            hooks = (5, 11, 17, 23) if hooks is None else hooks
            self.vit_features = 1024

        elif backbone == "vit_tiny":

            scratch_in_features = (48, 96, 144, 192)
            self.model = MODEL_REGISTRY.get("VanillaViT")(
                img_size=img_size,
                patch_size=16,
                embedding_dim=192,
                head_dim=64,
                depth=12,
                attn_heads=3,
                encoder_mlp_dim=192,
                n_classes=3,  # doenst matter because decoder part is not used in DPTs forward_vit
                in_channels=in_channels,
            )
            hooks = (2, 5, 8, 11) if hooks is None else hooks
            self.vit_features = 192

        else:
            raise NotImplementedError

        assert readout in (
            "add",
            "ignore",
            "project",
        ), f"Not valid `readout` param, Must be one of ('add','ignore','project'), but got {readout}"  # check if valid string
        features = scratch_in_features[0]

        self._register_hooks_and_add_postprocess(
            size=img_size,
            features=scratch_in_features,
            hooks=hooks,
            use_readout=readout,
            enable_attention_hooks=enable_attention_hooks,
            start_index=start_index,
        )
        self._make_scratch(
            in_shape=scratch_in_features,
            out_shape=features,
            groups=1,
            expand=False,
        )
        self._add_refinenet_to_scratch(features=features, use_bn=use_bn)

        self.model.head = nn.Sequential(
            nn.Conv2d(features, features // 2, kernel_size=3, stride=1, padding=1),
            Interpolate(scale_factor=2, mode="bilinear", align_corners=True),
            nn.Conv2d(features // 2, 32, kernel_size=3, stride=1, padding=1),
            nn.ReLU(True),
            nn.Conv2d(32, 1, kernel_size=1, stride=1, padding=0),
            nn.ReLU(True) if non_negative else nn.Identity(),
            nn.Identity(),
        )

    def _register_hooks_and_add_postprocess(
        self,
        size=(384, 384),
        features=(96, 192, 384, 768),
        hooks=(2, 5, 8, 11),
        use_readout="ignore",
        enable_attention_hooks=False,
        start_index=1,
    ):
        """
        Registers forward hooks to the backbone and initializes activation-postprocessing-blocks (act_postprocess(int))
        Parameters
        ----------
        size: tuple[int]
            Input image size
        features:tuple[int]
            Number of features
        hooks:tuple[int]
            List containing index of encoder blocks to which forward hooks will be registered
        use_readout:str
            Appropriate readout operation,must be one of  {`add`,`ignore`,`project`}
        enable_attention_hooks:bool
            If True, forward hooks will be registered to attention blocks.
        start_index:int
            Parameter that handles readout operation, default value is 1.
        """

        for i in range(4):
            self.model.encoder.encoder[hooks[i]][0].fn.register_forward_hook(
                get_activation(str(i + 1))
            )

        self.activations = activations

        if enable_attention_hooks:
            for i in range(4):
                self.model.encoder.encoder[hooks[i]][0].fn.register_forward_hook(
                    get_attention(f"attn_{str(i+1)}")
                )
            self.attention = attention

        readout_oper = get_readout_oper(
            self.vit_features, features, use_readout, start_index
        )

        # 32, 48, 136, 384

        self.act_postprocess1 = nn.Sequential(
            readout_oper[0],
            Transpose(1, 2),
            nn.Unflatten(2, torch.Size([size[0] // 16, size[1] // 16])),
            nn.Conv2d(
                in_channels=self.vit_features,
                out_channels=features[0],
                kernel_size=1,
                stride=1,
                padding=0,
            ),
            nn.ConvTranspose2d(
                in_channels=features[0],
                out_channels=features[0],
                kernel_size=4,
                stride=4,
                padding=0,
                bias=True,
                dilation=1,
                groups=1,
            ),
        )

        self.act_postprocess2 = nn.Sequential(
            readout_oper[1],
            Transpose(1, 2),
            nn.Unflatten(2, torch.Size([size[0] // 16, size[1] // 16])),
            nn.Conv2d(
                in_channels=self.vit_features,
                out_channels=features[1],
                kernel_size=1,
                stride=1,
                padding=0,
            ),
            nn.ConvTranspose2d(
                in_channels=features[1],
                out_channels=features[1],
                kernel_size=2,
                stride=2,
                padding=0,
                bias=True,
                dilation=1,
                groups=1,
            ),
        )

        self.act_postprocess3 = nn.Sequential(
            readout_oper[2],
            Transpose(1, 2),
            nn.Unflatten(2, torch.Size([size[0] // 16, size[1] // 16])),
            nn.Conv2d(
                in_channels=self.vit_features,
                out_channels=features[2],
                kernel_size=1,
                stride=1,
                padding=0,
            ),
        )

        self.act_postprocess4 = nn.Sequential(
            readout_oper[3],
            Transpose(1, 2),
            nn.Unflatten(2, torch.Size([size[0] // 16, size[1] // 16])),
            nn.Conv2d(
                in_channels=self.vit_features,
                out_channels=features[3],
                kernel_size=1,
                stride=1,
                padding=0,
            ),
            nn.Conv2d(
                in_channels=features[3],
                out_channels=features[3],
                kernel_size=3,
                stride=2,
                padding=1,
            ),
        )

        self.model.start_index = start_index
        self.model.patch_size = [16, 16]

        self.model.forward_flex = types.MethodType(forward_flex, self.model)
        self.model._resize_pos_embed = types.MethodType(_resize_pos_embed, self.model)

    def _make_scratch(self, in_shape, out_shape, groups=1, expand=False):
        """
        Makes a scratch module which is subclass of nn.Module

        Parameters
        ----------
        in_shape: list[int]
        out_shape:int
        groups: int
        expand:bool
        """
        self.scratch = nn.Module()

        for i in range(4):
            layer = nn.Conv2d(
                in_shape[i],
                out_shape * (2) ** (i) if expand else out_shape,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
                groups=groups,
            )
            setattr(self.scratch, f"layer{i+1}_rn", layer)

    def _add_refinenet_to_scratch(self, features, use_bn):
        """

        Parameters
        ----------
        features: int
            Number of features
        use_bn: bool
            Whether to use batch normalisation
        """
        for i in range(4):
            refinenet = FeatureFusionBlock_custom(
                features,
                nn.ReLU(False),
                deconv=False,
                bn=use_bn,
                expand=False,
                align_corners=True,
            )
            setattr(self.scratch, f"refinenet{i+1}", refinenet)

    def forward_vit(self, x):
        """
        Performs forward pass on backbone ViT model and fetches output from different encoder blocks with the help of hooks

        Parameters
        -----------
        x: torch.Tensor
            Input image tensor
        """

        b, c, h, w = x.shape

        glob = forward_flex(self, x)

        layer_1 = self.activations["1"]
        layer_2 = self.activations["2"]
        layer_3 = self.activations["3"]
        layer_4 = self.activations["4"]

        layer_1 = self.act_postprocess1[0:2](layer_1)
        layer_2 = self.act_postprocess2[0:2](layer_2)
        layer_3 = self.act_postprocess3[0:2](layer_3)
        layer_4 = self.act_postprocess4[0:2](layer_4)

        unflatten = nn.Sequential(
            nn.Unflatten(
                2,
                torch.Size(
                    [
                        h // self.model.patch_size[1],
                        w // self.model.patch_size[0],
                    ]
                ),
            )
        )

        if layer_1.ndim == 3:
            layer_1 = unflatten(layer_1)
        if layer_2.ndim == 3:
            layer_2 = unflatten(layer_2)
        if layer_3.ndim == 3:
            layer_3 = unflatten(layer_3)
        if layer_4.ndim == 3:
            layer_4 = unflatten(layer_4)

        layer_1 = self.act_postprocess1[3 : len(self.act_postprocess1)](layer_1)
        layer_2 = self.act_postprocess2[3 : len(self.act_postprocess2)](layer_2)
        layer_3 = self.act_postprocess3[3 : len(self.act_postprocess3)](layer_3)
        layer_4 = self.act_postprocess4[3 : len(self.act_postprocess4)](layer_4)

        return layer_1, layer_2, layer_3, layer_4

    def forward(self, x):
        """
        Forward pass of DPTDepth

        Parameters
        ----------
        x:torch.Tensor
            Input image tensor
        """
        if self.channels_last:
            x.contiguous(memory_format=torch.channels_last)

        layer_1, layer_2, layer_3, layer_4 = self.forward_vit(x)

        layer_1 = self.scratch.layer1_rn(layer_1)
        layer_2 = self.scratch.layer2_rn(layer_2)
        layer_3 = self.scratch.layer3_rn(layer_3)
        layer_4 = self.scratch.layer4_rn(layer_4)

        path1 = self.scratch.refinenet4(layer_4)
        path1 = self.scratch.refinenet3(path1, layer_3)
        path1 = self.scratch.refinenet2(path1, layer_2)
        path1 = self.scratch.refinenet1(path1, layer_1)

        inv_depth = self.model.head(path1).squeeze(dim=1)

        if self.invert:
            depth = self.scale * inv_depth + self.shift
            depth[depth < 1e-8] = 1e-8
            depth = 1.0 / depth
            return depth
        else:
            return inv_depth


class Slice(nn.Module):
    """Handles readout operation when `readout` parameter is `ignore`. Removes `cls_token` or  `readout_token` by index slicing"""

    def __init__(self, start_index=1):
        super(Slice, self).__init__()
        self.start_index = start_index

    def forward(self, x):
        return x[:, self.start_index :]


class AddReadout(nn.Module):
    """Handles readout operation when `readout` parameter is `add`. Removes `cls_token` or  `readout_token` from tensor and adds it to the rest of tensor"""

    def __init__(self, start_index=1):
        super(AddReadout, self).__init__()
        self.start_index = start_index

    def forward(self, x):

        readout = x[:, 0]
        return x[:, self.start_index :] + readout.unsqueeze(1)


class ProjectReadout(nn.Module):
    """Another class that handles readout operation. Used when `readout` parameter is `project`"""

    def __init__(self, in_features, start_index=1):
        super(ProjectReadout, self).__init__()
        self.start_index = start_index

        self.project = nn.Sequential(nn.Linear(2 * in_features, in_features), nn.GELU())

    def forward(self, x):
        readout = x[:, 0].unsqueeze(1).expand_as(x[:, self.start_index :])
        features = torch.cat((x[:, self.start_index :], readout), -1)

        return self.project(features)


class Interpolate(nn.Module):
    """Interpolation module

    Parameters
    ----------
    scale_factor : float
        Scaling factor used in interpolation
    mode :str
        Interpolation mode
    align_corners: bool
        Whether to align corners in Interpolation operation
    """

    def __init__(self, scale_factor, mode, align_corners=False):

        super(Interpolate, self).__init__()

        self.interp = nn.functional.interpolate
        self.scale_factor = scale_factor
        self.mode = mode
        self.align_corners = align_corners

    def forward(self, x):
        """Forward pass"""

        x = self.interp(
            x,
            scale_factor=self.scale_factor,
            mode=self.mode,
            align_corners=self.align_corners,
        )

        return x


class Transpose(nn.Module):
    def __init__(self, dim0, dim1):
        super(Transpose, self).__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def forward(self, x):
        x = x.transpose(self.dim0, self.dim1)
        return x


class ResidualConvUnit_custom(nn.Module):
    """Residual convolution module

    Parameters
    ----------
    features :int
        Number of features
    activation: nn.Module
        Activation module, default is nn.GELU
    bn: bool
        Whether to use batch normalisation
    """

    def __init__(self, features, activation=nn.GELU, bn=True):
        super().__init__()
        self.bn = bn
        self.groups = 1

        self.conv1 = nn.Conv2d(
            features,
            features,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=not self.bn,
            groups=self.groups,
        )

        self.conv2 = nn.Conv2d(
            features,
            features,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=not self.bn,
            groups=self.groups,
        )

        if self.bn == True:
            self.bn1 = nn.BatchNorm2d(features)
            self.bn2 = nn.BatchNorm2d(features)

        self.activation = activation

        self.skip_add = nn.quantized.FloatFunctional()

    def forward(self, x):
        """forward pass"""
        out = self.activation(x)
        out = self.conv1(out)
        if self.bn == True:
            out = self.bn1(out)

        out = self.activation(out)
        out = self.conv2(out)
        if self.bn == True:
            out = self.bn2(out)

        return self.skip_add.add(out, x)

        # return out + x


class FeatureFusionBlock_custom(nn.Module):
    """Feature fusion block."""

    def __init__(
        self,
        features,
        activation,
        deconv=False,
        bn=False,
        expand=False,
        align_corners=True,
    ):

        super(FeatureFusionBlock_custom, self).__init__()

        self.deconv = deconv
        self.align_corners = align_corners

        self.groups = 1

        self.expand = expand
        out_features = features

        self.out_conv = nn.Conv2d(
            features,
            out_features,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=True,
            groups=1,
        )

        self.resConfUnit1 = ResidualConvUnit_custom(features, activation, bn)
        self.resConfUnit2 = ResidualConvUnit_custom(features, activation, bn)

        self.skip_add = nn.quantized.FloatFunctional()

    def forward(self, *xs):
        """Forward pass"""
        output = xs[0]

        if len(xs) == 2:
            res = self.resConfUnit1(xs[1])
            output = self.skip_add.add(output, res)
            # output += res

        output = self.resConfUnit2(output)

        output = nn.functional.interpolate(
            output, scale_factor=2, mode="bilinear", align_corners=self.align_corners
        )

        output = self.out_conv(output)

        return output


def get_readout_oper(vit_features, features, use_readout, start_index=1):
    if use_readout == "ignore":
        readout_oper = [Slice(start_index)] * len(features)
    elif use_readout == "add":
        readout_oper = [AddReadout(start_index)] * len(features)
    else:
        readout_oper = [
            ProjectReadout(vit_features, start_index) for out_feat in features
        ]
    return readout_oper



import torch
import torch.nn as nn
from einops import repeat

from ...common import BaseClassificationModel
from ...decoder import MLPDecoder
from ...encoder import CrossEncoder, LinearEmbedding
from ...utils import MODEL_REGISTRY


class _cross_p(BaseClassificationModel):
    def __init__(
        self,
        img_size,
        patch_size,
        latent_dim=1024,
        in_channels=3,
        p_dropout_embedding=0.0,
    ):
        super().__init__(img_size, patch_size, in_channels)

        self.patch_embedding = LinearEmbedding(
            latent_dim, self.patch_height, self.patch_width, self.patch_dim
        )

        self.pos_embedding = nn.Parameter(
            torch.randn(1, self.num_patches + 1, latent_dim)
        )
        self.cls_token = nn.Parameter(torch.randn(1, 1, latent_dim))
        self.embedding_dropout = nn.Dropout(p_dropout_embedding)

    def forward(self, x):

        x = self.patch_embedding(x)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, "() n d -> b n d", b=b)
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding[:, : (n + 1)]
        x = self.embedding_dropout(x)

        return x


@MODEL_REGISTRY.register()
class CrossViT(BaseClassificationModel):
    """
    Implementation of 'CrossViT: Cross-Attention Multi-Scale Vision Transformer for Image Classification'
    https://arxiv.org/abs/2103.14899

    Parameters
    ----------
    img_size: int
        Size of the image
    patch_size_s: int
        Size of the smaller patches
    patch_size_l: int
        Size of the larger patches
    n_classes: int
        Number of classes for classification
    cross_dim_head_s: int
        Dimension of the head of the cross-attention for the smaller patches
    cross_dim_head_l: int
        Dimension of the head of the cross-attention for the larger patches
    latent_dim_s: int
        Dimension of the hidden layer for the smaller patches
    latent_dim_l: int
        Dimension of the hidden layer for the larger patches
    head_dim_s: int
        Dimension of the head of the attention for the smaller patches
    head_dim_l: int
        Dimension of the head of the attention for the larger patches
    depth_s: int
        Number of attention layers in encoder for the smaller patches
    depth_l: int
        Number of attention layers in encoder for the larger patches
    attn_heads_s: int
        Number of attention heads for the smaller patches
    attn_heads_l: int
        Number of attention heads for the larger patches
    cross_head_s: int
        Number of CrossAttention heads for the smaller patches
    cross_head_l: int
        Number of CrossAttention heads for the larger patches
    encoder_mlp_dim_s: int
        Dimension of hidden layer in the encoder for the smaller patches
    encoder_mlp_dim_l: int
        Dimension of hidden layer in the encoder for the larger patches
    in_channels: int
        Number of input channels
    decoder_config_s: int or tuple or list, optional
        Configuration of the decoder for the smaller patches
    decoder_config_l: int or tuple or list, optional
        Configuration of the decoder for the larger patches
    pool_s: {"cls","mean"}
        Feature pooling type for the smaller patches
    pool_l: {"cls","mean"}
        Feature pooling type for the larger patches
    p_dropout_encoder_s: float
        Dropout probability in the encoder for the smaller patches
    p_dropout_encoder_l: float
        Dropout probability in the encoder for the larger patches
    p_dropout_embedding_s: float
        Dropout probability in the embedding layer for the smaller patches
    p_dropout_embedding_l: float
        Dropout probability in the embedding layer for the larger patches
    """

    def __init__(
        self,
        img_size,
        patch_size_s,
        patch_size_l,
        n_classes,
        cross_dim_head_s=64,
        cross_dim_head_l=64,
        latent_dim_s=1024,
        latent_dim_l=1024,
        head_dim_s=64,
        head_dim_l=64,
        depth_s=6,
        depth_l=6,
        attn_heads_s=16,
        attn_heads_l=16,
        cross_head_s=8,
        cross_head_l=8,
        encoder_mlp_dim_s=2048,
        encoder_mlp_dim_l=2048,
        in_channels=3,
        decoder_config_s=None,
        decoder_config_l=None,
        pool_s="cls",
        pool_l="cls",
        p_dropout_encoder_s=0.0,
        p_dropout_encoder_l=0.0,
        p_dropout_embedding_s=0.0,
        p_dropout_embedding_l=0.0,
    ):
        super().__init__(img_size, patch_size_s, in_channels, pool_s)
        super().__init__(img_size, patch_size_l, in_channels, pool_l)

        self.s = _cross_p(
            img_size, patch_size_s, latent_dim_s, in_channels, p_dropout_embedding_s
        )
        self.l = _cross_p(
            img_size, patch_size_l, latent_dim_l, in_channels, p_dropout_embedding_l
        )
        self.encoder = CrossEncoder(
            latent_dim_s,
            latent_dim_l,
            head_dim_s,
            head_dim_l,
            cross_dim_head_s,
            cross_dim_head_l,
            depth_s,
            depth_l,
            attn_heads_s,
            attn_heads_l,
            cross_head_s,
            cross_head_l,
            encoder_mlp_dim_s,
            encoder_mlp_dim_l,
            p_dropout_encoder_s,
            p_dropout_encoder_l,
        )
        self.pool_s = lambda x: x.mean(dim=1) if pool_s == "mean" else x[:, 0]
        self.pool_l = lambda x: x.mean(dim=1) if pool_l == "mean" else x[:, 0]

        if decoder_config_s is not None:

            if not isinstance(decoder_config_s, list):
                decoder_config_s = list(decoder_config_s)

            assert (
                decoder_config_s[0] == latent_dim_s
            ), "`latent_dim` should be equal to the first item of `decoder_config`"

            self.decoder_s = MLPDecoder(decoder_config_s, n_classes)

        else:
            self.decoder_s = MLPDecoder(latent_dim_s, n_classes)

        if decoder_config_l is not None:

            if not isinstance(decoder_config_l, list):
                decoder_config_l = list(decoder_config_l)

            assert (
                decoder_config_l[0] == latent_dim_l
            ), "`latent_dim` should be equal to the first item of `decoder_config`"

            self.decoder_l = MLPDecoder(decoder_config_l, n_classes)

        else:
            self.decoder_l = MLPDecoder(latent_dim_l, n_classes)

    def forward(self, img):
        """

        Parameters
        ----------
        img: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of size `n_classes`

        """
        emb_s = self.s(img)
        emb_l = self.l(img)
        emb_s, emb_l = self.encoder(emb_s, emb_l)
        cls_s = self.pool_s(emb_s)
        cls_l = self.pool_l(emb_l)
        n_s = self.decoder_s(cls_s)
        n_l = self.decoder_l(cls_l)
        n = n_s + n_l

        return n



import torch
import torch.nn as nn
from timm.models.layers import trunc_normal_

from ...encoder.convvt import ConvVTStage
from ...utils import MODEL_REGISTRY


@MODEL_REGISTRY.register()
class ConvVT(nn.Module):
    """
    Implementation of CvT: Introducing Convolutions to Vision Transformers:
    https://arxiv.org/pdf/2103.15808.pdf

    Parameters:
    ------------
    img_size: int
        Size of the image, default is 224
    in_channels:int
        Number of input channels in image, default is 3
    num_stages: int
        Number of stages in encoder block, default is 3
    n_classes: int
        Number of classes for classification, default is 1000
    * The following are all in list of int/float with length num_stages
    patch_size: list[int]
        Size of patch, default is [7, 3, 3]
    patch_stride: list[int]
        Stride of patch, default is [4, 2, 2]
    patch_padding: list[int]
        Padding for patch, default is [2, 1, 1]
    embedding_dim: list[int]
        Embedding dimensions, default is [64, 192, 384]
    depth: list[int]
        Number of CVT Attention blocks in each stage, default is [1, 2, 10]
    num_heads: list[int]
        Number of heads in attention, default is [1, 3, 6]
    mlp_ratio: list[float]
        Feature dimension expansion ratio in MLP, default is [4.0, 4.0, 4.0]
    p_dropout: list[float]
        Probability of dropout in MLP, default is [0, 0, 0]
    attn_dropout: list[float]
        Probability of dropout in attention, default is [0, 0, 0]
    drop_path_rate: list[float]
        Probability for droppath, default is [0, 0, 0.1]
    kernel_size: list[int]
        Size of kernel, default is [3, 3, 3]
    padding_q: list[int]
        Size of padding in q, default is [1, 1, 1]
    padding_kv: list[int]
        Size of padding in kv, default is [2, 2, 2]
    stride_kv: list[int]
        Stride in kv, default is [2, 2, 2]
    stride_q: list[int]
        Stride in q, default is [1, 1, 1]
    """

    def __init__(
        self,
        img_size=224,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        embedding_dim=[64, 192, 384],
        num_heads=[1, 3, 6],
        depth=[1, 2, 10],
        mlp_ratio=[4.0, 4.0, 4.0],
        p_dropout=[0, 0, 0],
        attn_dropout=[0, 0, 0],
        drop_path_rate=[0, 0, 0.1],
        kernel_size=[3, 3, 3],
        padding_q=[1, 1, 1],
        padding_kv=[1, 1, 1],
        stride_kv=[2, 2, 2],
        stride_q=[1, 1, 1],
        in_channels=3,
        num_stages=3,
        n_classes=1000,
    ):
        super().__init__()

        self.n_classes = n_classes

        self.num_stages = num_stages
        self.stages = nn.ModuleList([])
        for i in range(self.num_stages):
            stage = ConvVTStage(
                in_channels=in_channels,
                img_size=img_size // (4 * 2 ** i),
                with_cls_token=False if i < self.num_stages - 1 else True,
                patch_size=patch_size[i],
                patch_stride=patch_stride[i],
                patch_padding=patch_padding[i],
                embedding_dim=embedding_dim[i],
                depth=depth[i],
                num_heads=num_heads[i],
                mlp_ratio=mlp_ratio[i],
                p_dropout=p_dropout[i],
                attn_dropout=attn_dropout[i],
                drop_path_rate=drop_path_rate[i],
                kernel_size=kernel_size[i],
                padding_q=padding_q[i],
                padding_kv=padding_kv[i],
                stride_kv=stride_kv[i],
                stride_q=stride_q[i],
            )
            self.stages.append(stage)
            in_channels = embedding_dim[i]

        self.norm = nn.LayerNorm(embedding_dim[-1])

        # Classifier head
        self.head = (
            nn.Linear(embedding_dim[-1], n_classes) if n_classes > 0 else nn.Identity()
        )
        trunc_normal_(self.head.weight, std=0.02)

    def forward(self, x):

        for i in range(self.num_stages):
            x, cls_tokens = self.stages[i](x)

        x = self.norm(cls_tokens)
        x = torch.squeeze(x)
        x = self.head(x)

        return x



import torch.nn as nn

from ...utils import pair


class PatchEmbedding(nn.Module):
    """

    Parameters
    ----------
    img_size: int
        Image Size
    patch_size: int
        Patch Size
    in_channels: int
        Number of input channels in the image
    embedding_dim: int
        Number of linear projection output channels
    norm_layer: nn.Module,
        Normalization layer, Default is `nn.LayerNorm`

    """

    def __init__(
        self,
        img_size,
        patch_size,
        in_channels,
        embedding_dim,
        norm_layer=nn.LayerNorm,
    ):
        super(PatchEmbedding, self).__init__()

        self.img_size = pair(img_size)
        self.patch_size = pair(patch_size)
        self.patch_resolution = [
            self.img_size[0] // self.patch_size[0],
            self.img_size[1] // self.patch_size[1],
        ]

        self.proj = nn.Conv2d(
            in_channels=in_channels,
            out_channels=embedding_dim,
            kernel_size=patch_size,
            stride=patch_size,
        )
        self.norm = norm_layer(embedding_dim)

    def forward(self, x):
        """

        Parameters
        ----------
        x:torch.Tensor
            Input tensor

        Returns
        ----------
        torch.Tensor
            Returns output tensor by applying convolution operation with same `kernel_size` and `stride` on input tensor.

        """
        B, C, H, W = x.shape

        assert (
            H == self.img_size[0] and W == self.img_size[1]
        ), f"Input Image Size {H}*{W} doesnt match model {self.img_size[0]}*{self.img_size[1]}"

        x = self.proj(x).flatten(2).transpose(1, 2)
        x = self.norm(x)

        return x



import torch
import torch.nn as nn
from timm.models.layers import trunc_normal_

from ...common import BaseClassificationModel
from ...decoder import MLPDecoder
from ...encoder import PatchEmbedding, PosEmbedding, SwinEncoder
from ...functional import PatchMerging
from ...utils import MODEL_REGISTRY


@MODEL_REGISTRY.register()
class SwinTransformer(BaseClassificationModel):
    """
    Implementation of `Swin Transformer: Hierarchical Vision Transformer using Shifted Windows`
    https://arxiv.org/abs/2103.14030v1

    Parameters
    ----------
    img_size: int
        Size of an Image
    patch_size: int
        Patch Size
    in_channels: int
        Input channels in image, default=3
    n_classes: int
        Number of classes for classification
    embedding_dim: int
        Patch Embedding dimension
    depths: tuple[int]
        Depth in each Transformer layer
    num_heads: tuple[int]
        Number of heads in each transformer layer
    window_size: int
        Window Size
    mlp_ratio : float
        Ratio of mlp heads to embedding dimension
    qkv_bias: bool, default= True
        Adds bias to the qkv if true
    qk_scale:  float, optional
        Override default qk scale of head_dim ** -0.5 in Window Attention if set
    p_dropout: float
        Dropout rate, default is 0.0
    attn_dropout: float
        Attention dropout rate,default is 0.0
    drop_path_rate: float
        Stochastic depth rate, default is 0.1
    norm_layer: nn.Module
        Normalization layer,default is nn.LayerNorm
    ape: bool, optional
        Whether to add relative/absolute position embedding to patch embedding, default is True
    decoder_config: int or tuple[int], optional
        Configuration of the decoder. If None, the default configuration is used.
    patch_norm: bool, optional
        Whether to add Normalization layer in PatchEmbedding, default is True
    """

    def __init__(
        self,
        img_size,
        patch_size,
        in_channels,
        n_classes,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=8,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path_rate=0.1,
        norm_layer=nn.LayerNorm,
        ape=True,
        decoder_config=None,
        patch_norm=True,
    ):
        super(SwinTransformer, self).__init__(
            img_size, patch_size, in_channels, pool="mean"
        )

        self.patch_embed = PatchEmbedding(
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            embedding_dim=embedding_dim,
            norm_layer=norm_layer if patch_norm else nn.Identity,
        )
        self.patch_resolution = self.patch_embed.patch_resolution
        num_patches = self.patch_resolution[0] * self.patch_resolution[1]
        self.ape = ape
        num_features = int(embedding_dim * 2 ** (len(depths) - 1))

        self.absolute_pos_embed = (
            PosEmbedding(shape=num_patches, dim=embedding_dim, drop=p_dropout, std=0.02)
            if ape
            else nn.Identity()
        )
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        self.encoder = nn.ModuleList()

        for i_layer in range(len(depths)):
            layer = SwinEncoder(
                dim=int(embedding_dim * (2 ** i_layer)),
                input_resolution=(
                    (self.patch_resolution[0] // (2 ** i_layer)),
                    self.patch_resolution[1] // (2 ** i_layer),
                ),
                depth=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qkv_scale=qk_scale,
                p_dropout=p_dropout,
                attn_dropout=attn_dropout,
                drop_path=dpr[sum(depths[:i_layer]) : sum(depths[: i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging if i_layer < len(depths) - 1 else None,
            )
            self.encoder.append(layer)

        if decoder_config is not None:

            if not isinstance(decoder_config, list):
                decoder_config = list(decoder_config)

            assert (
                decoder_config[0] == num_features
            ), f"first item of `decoder_config` should be equal to the `num_features`; num_features=embed_dim * 2** (len(depths)-1) which is = {num_features} "

            self.decoder = MLPDecoder(decoder_config, n_classes)

        else:
            self.decoder = MLPDecoder(num_features, n_classes)

        self.pool = nn.AdaptiveAvgPool1d(1)
        self.norm = norm_layer(num_features) if norm_layer is not None else nn.Identity
        self.pos_drop = nn.Dropout(p=p_dropout)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of size `n_classes`

        """
        x = self.patch_embed(x)

        x = self.absolute_pos_embed(x)

        for layer in self.encoder:
            x = layer(x)

        x = self.norm(x)
        x = self.pool(x.transpose(1, 2)).flatten(1)
        x = self.decoder(x)

        return x



import torch.nn as nn


class FeedForward(nn.Module):
    """

    Parameters
    ----------
    dim: int
        Dimension of the input tensor
    hidden_dim: int, optional
        Dimension of hidden layer
    out_dim: int, optional
        Dimension of the output tensor
    p_dropout: float
        Dropout probability, default=0.0

    """

    def __init__(self, dim, hidden_dim=None, out_dim=None, p_dropout=0.0):
        super().__init__()

        out_dim = out_dim if out_dim is not None else dim
        hidden_dim = hidden_dim if hidden_dim is not None else dim

        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(p_dropout),
            nn.Linear(hidden_dim, out_dim),
            nn.Dropout(p_dropout),
        )

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------

        torch.Tensor
            Returns output tensor by performing linear operations and activation on input tensor

        """

        return self.net(x)



import torch
import torch.nn as nn

from vformer.encoder.embedding import TubeletEmbedding
from vformer.functional import PatchMerging
from vformer.utils import ENCODER_REGISTRY

encoder_modules = ENCODER_REGISTRY.get_list()


def test_VanillaEncoder():

    test_tensor = torch.randn(2, 65, 1024)
    encoder = ENCODER_REGISTRY.get("VanillaEncoder")(
        embedding_dim=1024, depth=6, num_heads=16, head_dim=64, mlp_dim=2048
    )
    out = encoder(test_tensor)
    assert out.shape == test_tensor.shape  # shape remains same
    del encoder, test_tensor


def test_SwinEncoder():

    test_tensor = torch.randn(3, 3136, 96)

    # when downsampled
    encoder = ENCODER_REGISTRY.get("SwinEncoder")(
        dim=96,
        input_resolution=(224 // 4, 224 // 4),
        depth=2,
        num_heads=3,
        window_size=7,
        downsample=PatchMerging,
    )
    out = encoder(test_tensor)

    assert out.shape == (3, 784, 192)
    del encoder

    # when not downsampled
    encoder = ENCODER_REGISTRY.get("SwinEncoder")(
        dim=96,
        input_resolution=(224 // 4, 224 // 4),
        depth=2,
        num_heads=3,
        window_size=7,
        downsample=None,
        use_checkpoint=True,
    )
    out = encoder(test_tensor)
    assert out.shape == (3, 3136, 96)
    del encoder

    encoder_block = ENCODER_REGISTRY.get("SwinEncoderBlock")(
        dim=96, input_resolution=(224 // 4, 224 // 4), num_heads=3, window_size=7
    )
    out = encoder_block(test_tensor)
    assert out.shape == test_tensor.shape
    del encoder_block


def test_PVTEncoder():

    test_tensor = torch.randn(4, 3136, 64)

    encoder = ENCODER_REGISTRY.get("PVTEncoder")(
        dim=64,
        depth=3,
        qkv_bias=True,
        qk_scale=0.0,
        p_dropout=0.0,
        attn_dropout=0.1,
        drop_path=[0.0] * 3,
        act_layer=nn.GELU,
        sr_ratio=1,
        linear=False,
        use_dwconv=False,
        num_heads=1,
        mlp_ratio=4,
    )
    out = encoder(test_tensor, H=56, W=56)
    assert out.shape == test_tensor.shape
    del encoder


def test_CrossEncoder():

    test_tensor1 = torch.randn(3, 5, 128)
    test_tensor2 = torch.randn(3, 5, 256)

    encoder = ENCODER_REGISTRY.get("CrossEncoder")(128, 256)
    out = encoder(test_tensor1, test_tensor2)
    assert out[0].shape == test_tensor1.shape
    assert out[1].shape == test_tensor2.shape
    del encoder


def test_ConViTEncoder():

    test_tensor = torch.randn(2, 64, 1024)
    encoder = ENCODER_REGISTRY.get("ConViTEncoder")(
        embedding_dim=1024, depth=6, num_heads=16, head_dim=64, mlp_dim=2048
    )
    out = encoder(test_tensor)
    assert out.shape == test_tensor.shape  # shape remains same
    del encoder, test_tensor


def test_ConvVTStage():
    test_tensor1 = torch.randn(32, 3, 224, 224)

    encoder = ENCODER_REGISTRY.get("ConvVTStage")(
        patch_size=7,
        patch_stride=4,
        patch_padding=2,
        img_size=56,
        embedding_dim=64,
        depth=1,
        num_heads=1,
        with_cls_token=True,
    )
    out, cls_tokens = encoder(test_tensor1)
    assert out.shape == torch.Size([32, 64, 56, 56])
    del encoder

    test_tensor1 = torch.randn(32, 3, 224, 224)

    encoder = ENCODER_REGISTRY.get("ConvVTStage")(
        patch_size=7,
        patch_stride=4,
        patch_padding=2,
        img_size=56,
        embedding_dim=64,
        depth=1,
        num_heads=1,
        with_cls_token=True,
        init="xavier",
    )
    out, cls_tokens = encoder(test_tensor1)
    assert out.shape == torch.Size([32, 64, 56, 56])
    del encoder


def test_TubeletEmbedding():
    test_tensor = torch.randn(
        7, 20, 3, 224, 224
    )  # batch_size,time,in_channels,height,width
    embedding = TubeletEmbedding(
        embedding_dim=192, tubelet_w=16, tubelet_t=5, tubelet_h=16, in_channels=3
    )
    out = embedding(test_tensor)
    assert out.shape == (
        7,
        4,
        196,
        192,
    )  # batch,time/tubelet_t,height*width/(tubelet_h,tubelet_w),embeeding_dim
    del embedding

    test_tensor = torch.randn(11, 15, 1, 28, 28)
    embedding = TubeletEmbedding(96, 5, 7, 7, 1)
    out = embedding(test_tensor)
    assert out.shape == (11, 3, 16, 96)
    del embedding


def test_ViViTEncoder():

    encoder = ENCODER_REGISTRY.get("ViViTEncoder")(
        dim=192, num_heads=3, head_dim=64, p_dropout=0.0, depth=3
    )

    test_tensor = torch.randn(7, 20, 196, 192)
    logits = encoder(test_tensor)
    assert logits.shape == (7, 3920, 192)



import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.models.layers import trunc_normal_

from ...utils import pair


class PVTPosEmbedding(nn.Module):
    """

    Parameters
    ----------
    pos_shape : int or tuple(int)
        The shape of the absolute position embedding.
    pos_dim : int
        The dimension of the absolute position embedding.
    p_dropout : float, optional
        Probability of an element to be zeroed, default is 0.2
    std: float
        Standard deviation for truncated normal distribution
    """

    def __init__(self, pos_shape, pos_dim, p_dropout=0.0, std=0.02):
        super().__init__()

        pos_shape = pair(pos_shape)
        self.pos_embed = nn.Parameter(
            torch.zeros(1, pos_shape[0] * pos_shape[1], pos_dim)
        )
        self.pos_shape = pos_shape
        self.pos_dim = pos_dim

        self.drop = nn.Dropout(p=p_dropout)
        trunc_normal_(self.pos_embed, std=std)

    def resize_pos_embed(self, pos_embed, shape, mode="bilinear", **kwargs):
        """
        Parameters
        ----------
            pos_embed : torch.Tensor
                Position embedding weights
            shape : tuple
                Required shape
            mode : str  ('nearest' | 'linear' | 'bilinear' | 'bicubic' | 'trilinear')
                Algorithm used for up/down sampling, default is 'bilinear'
        """
        assert pos_embed.ndim == 3, "shape of pos_embed must be [B, L, C]"

        pos_h, pos_w = self.pos_shape
        pos_embed_weight = pos_embed[:, (-1 * pos_h * pos_w) :]
        pos_embed_weight = (
            pos_embed_weight.reshape(1, pos_h, pos_w, self.pos_dim)
            .permute(0, 3, 1, 2)
            .contiguous()
        )
        pos_embed_weight = F.interpolate(
            pos_embed_weight, size=shape, mode=mode, **kwargs
        )
        pos_embed_weight = (
            torch.flatten(pos_embed_weight, 2).transpose(1, 2).contiguous()
        )
        pos_embed = pos_embed_weight

        return pos_embed

    def forward(self, x, H, W, mode="bilinear"):
        try:
            x = x + self.pos_embed

        except:
            x = x + self.resize_pos_embed(self.pos_embed, (H, W), mode)

        return self.drop(x)


class PosEmbedding(nn.Module):
    def __init__(self, shape, dim, drop=None, sinusoidal=False, std=0.02):
        super(PosEmbedding, self).__init__()

        if not sinusoidal:
            if isinstance(shape, int):
                shape = [1, shape, dim]
            else:
                shape = [1] + list(shape) + [dim]
            self.pos_embed = nn.Parameter(torch.zeros(shape))

        else:
            pe = torch.FloatTensor(
                [
                    [p / (10000 ** (2 * (i // 2) / dim)) for i in range(dim)]
                    for p in range(shape)
                ]
            )
            pe[:, 0::2] = torch.sin(pe[:, 0::2])
            pe[:, 1::2] = torch.cos(pe[:, 1::2])
            self.pos_embed = pe
            self.pos_embed.requires_grad = False
        trunc_normal_(self.pos_embed, std=std)
        self.pos_drop = nn.Dropout(drop) if drop is not None else nn.Identity()

    def forward(self, x):
        x = x + self.pos_embed
        return self.pos_drop(x)



import torch
import torch.nn as nn
from einops import repeat

from ...common import BaseClassificationModel
from ...decoder import MLPDecoder
from ...encoder import LinearEmbedding, PosEmbedding, VanillaEncoder
from ...utils import MODEL_REGISTRY


@MODEL_REGISTRY.register()
class VanillaViT(BaseClassificationModel):
    """
    Implementation of 'An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale'
    https://arxiv.org/abs/2010.11929

    Parameters
    ----------
    img_size: int
        Size of the image
    patch_size: int
        Size of a patch
    n_classes: int
        Number of classes for classification
    embedding_dim: int
        Dimension of hidden layer
    head_dim: int
        Dimension of the attention head
    depth: int
        Number of attention layers in the encoder
    attn_heads:int
        Number of the attention heads
    encoder_mlp_dim: int
        Dimension of hidden layer in the encoder
    in_channels: int
        Number of input channels
    decoder_config: int or tuple or list, optional
        Configuration of the decoder. If None, the default configuration is used.
    pool: {"cls","mean"}
        Feature pooling type
    p_dropout_encoder: float
        Dropout probability in the encoder
    p_dropout_embedding: float
        Dropout probability in the embedding layer
    """

    def __init__(
        self,
        img_size,
        patch_size,
        n_classes,
        embedding_dim=1024,
        head_dim=64,
        depth=6,
        attn_heads=16,
        encoder_mlp_dim=2048,
        in_channels=3,
        decoder_config=None,
        pool="cls",
        p_dropout_encoder=0.0,
        p_dropout_embedding=0.0,
    ):
        super().__init__(img_size, patch_size, in_channels, pool)

        self.patch_embedding = LinearEmbedding(
            embedding_dim, self.patch_height, self.patch_width, self.patch_dim
        )

        self.pos_embedding = PosEmbedding(
            shape=self.num_patches + 1,
            dim=embedding_dim,
            drop=p_dropout_embedding,
            sinusoidal=False,
        )
        self.cls_token = nn.Parameter(torch.randn(1, 1, embedding_dim))

        self.encoder = VanillaEncoder(
            embedding_dim=embedding_dim,
            depth=depth,
            num_heads=attn_heads,
            head_dim=head_dim,
            mlp_dim=encoder_mlp_dim,
            p_dropout=p_dropout_encoder,
        )
        self.pool = lambda x: x.mean(dim=1) if pool == "mean" else x[:, 0]

        if decoder_config is not None:

            if not isinstance(decoder_config, list):
                decoder_config = list(decoder_config)

            assert (
                decoder_config[0] == embedding_dim
            ), "`embedding_dim` should be equal to the first item of `decoder_config`"

            self.decoder = MLPDecoder(decoder_config, n_classes)

        else:
            self.decoder = MLPDecoder(embedding_dim, n_classes)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of size `n_classes`

        """
        x = self.patch_embedding(x)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, "() n d -> b n d", b=b)
        x = torch.cat((cls_tokens, x), dim=1)
        x = self.pos_embedding(x)
        x = self.encoder(x)
        x = self.pool(x)
        x = self.decoder(x)

        return x



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange

from ..utils import ATTENTION_REGISTRY
from .vanilla import VanillaSelfAttention


@ATTENTION_REGISTRY.register()
class GatedPositionalSelfAttention(VanillaSelfAttention):
    """
    Implementation of the Gated Positional Self-Attention from the paper: "ConViT: Improving Vision Transformers with Soft Convolutional Inductive Biases"

    Parameters
    ----------
    dim: int
        Dimension of the embedding
    num_heads: int
        Number of the attention heads, default is 8
    head_dim: int
        Dimension of each head, default is 64
    p_dropout: float
        Dropout probability, default is 0.0
    """

    def __init__(self, dim, num_heads=8, head_dim=64, p_dropout=0):
        super().__init__(dim, num_heads, head_dim, p_dropout)

        self.gate = nn.Parameter(torch.ones([1, num_heads, 1, 1]))
        self.pos = nn.Linear(3, num_heads)
        self.n_prev = 0

    def rel_embedding(self, n):

        l = int(n ** 0.5)
        rel_indices_x = torch.arange(l).reshape(1, -1)
        rel_indices_y = torch.arange(l).reshape(-1, 1)
        indices = rel_indices_x - rel_indices_y
        rel_indices_x = indices.repeat(l, l)
        rel_indices_y = indices.repeat_interleave(l, dim=0).repeat_interleave(l, dim=1)
        rel_indices_d = (rel_indices_x ** 2 + rel_indices_y ** 2) ** 0.5
        self.rel_indices = torch.stack(
            [rel_indices_x, rel_indices_y, rel_indices_d], dim=-1
        )
        self.rel_indices = self.rel_indices.unsqueeze(0)
        self.rel_indices = self.rel_indices.to(self.gate.device)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns output tensor by applying self-attention on input tensor

        """
        _, N, _ = x.shape

        if not self.n_prev == N:
            self.n_prev = N
            self.rel_embedding(N)
        qkv = self.to_qkv(x).chunk(3, dim=-1)

        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.num_heads), qkv
        )
        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale
        pos_attn = self.pos(self.rel_indices)
        attn = self.attend(dots)

        pos_attn = rearrange(pos_attn, "b n d h -> b h n d")
        pos_attn = F.softmax(pos_attn, dim=-1)
        attn = (1 - torch.sigmoid(self.gate)) * attn + torch.sigmoid(
            self.gate
        ) * pos_attn

        out = torch.matmul(attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")

        return self.to_out(out)



import numpy as np
import torch
import torch.nn as nn

from ....decoder import SegmentationHead
from ....encoder import OverlapPatchEmbed, PVTEncoder, PVTPosEmbedding
from ....utils import MODEL_REGISTRY


@MODEL_REGISTRY.register()
class PVTSegmentation(nn.Module):
    """
    Implementation of Pyramid Vision Transformer:
    https://arxiv.org/abs/2102.12122v1

    Parameters
    ----------
    img_size: int
        Image size
    patch_size: list(int)
        List of patch size
    in_channels: int
        Input channels in image, default=3
    embedding_dims:  int
        Patch Embedding dimension
    num_heads:tuple[int]
        Number of heads in each transformer layer
    depths: tuple[int]
        Depth in each Transformer layer
    mlp_ratio: float
        Ratio of mlp heads to embedding dimension
    qkv_bias: bool, default= True
        Adds bias to the qkv if true
    qk_scale: float, optional
        Override default qk scale of head_dim ** -0.5 in Spatial Attention if set
    p_dropout: float
        Dropout rate,default is 0.0
    attn_dropout:  float
        Attention dropout rate, default is 0.0
    drop_path_rate: float
        Stochastic depth rate, default is 0.1
    sr_ratios: float
        Spatial reduction ratio
    linear: bool
        Whether to use linear spatial attention
    use_dwconv: bool
        Whether to use Depth-wise convolutions in Overlap-patch embedding
    ape: bool
        Whether to use absolute position embedding
    return_pyramid:bool
        Whether to use all pyramid feature layers for up-sampling, default is False
    """

    def __init__(
        self,
        img_size=224,
        patch_size=[7, 3, 3, 3],
        in_channels=3,
        embedding_dims=[64, 128, 256, 512],
        num_heads=[1, 2, 4, 8],
        mlp_ratio=[4, 4, 4, 4],
        qkv_bias=False,
        qk_scale=None,
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        linear=False,
        out_channels=1,
        use_dwconv=False,
        ape=True,
        return_pyramid=False,
    ):
        super(PVTSegmentation, self).__init__()

        self.ape = ape
        self.depths = depths
        self.return_pyramid = return_pyramid

        assert (
            len(depths) == len(num_heads) == len(embedding_dims)
        ), "Configurations do not match"

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        self.patch_embeds = nn.ModuleList([])
        self.blocks = nn.ModuleList([])
        self.norms = nn.ModuleList()
        self.pos_embeds = nn.ModuleList()

        for i in range(len(depths)):

            self.patch_embeds.append(
                nn.ModuleList(
                    [
                        OverlapPatchEmbed(
                            img_size=img_size if i == 0 else img_size // (2 ** (i + 1)),
                            patch_size=patch_size[i],
                            stride=4 if i == 0 else 2,
                            in_channels=in_channels
                            if i == 0
                            else embedding_dims[i - 1],
                            embedding_dim=embedding_dims[i],
                        )
                    ]
                )
            )

            if ape:
                self.pos_embeds.append(
                    nn.ModuleList(
                        [
                            PVTPosEmbedding(
                                pos_shape=img_size // np.prod(patch_size[: i + 1]),
                                pos_dim=embedding_dims[i],
                            )
                        ]
                    )
                )

            self.blocks.append(
                nn.ModuleList(
                    [
                        PVTEncoder(
                            dim=embedding_dims[i],
                            num_heads=num_heads[i],
                            mlp_ratio=mlp_ratio[i],
                            qkv_bias=qkv_bias,
                            qk_scale=qk_scale,
                            p_dropout=p_dropout,
                            depth=depths[i],
                            attn_dropout=attn_dropout,
                            drop_path=dpr[sum(depths[:i]) : sum(depths[: i + 1])],
                            sr_ratio=sr_ratios[i],
                            linear=linear,
                            act_layer=nn.GELU,
                            use_dwconv=use_dwconv,
                        )
                    ]
                )
            )
            self.norms.append(norm_layer(embedding_dims[i]))

        self.head = SegmentationHead(
            out_channels=out_channels,
            embed_dims=embedding_dims if not return_pyramid else [embedding_dims[-1]],
        )

    def forward(self, x):

        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns output tensor
        """
        B = x.shape[0]
        out = []

        for i in range(len(self.depths)):

            patch_embed = self.patch_embeds[i]
            block = self.blocks[i]
            norm = self.norms[i]

            x, H, W = patch_embed[0](x)

            if self.ape:
                pos_embed = self.pos_embeds[i]
                x = pos_embed[0](x, H=H, W=W)

            for blk in block:
                x = blk(x, H=H, W=W)

            x = norm(x)
            x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
            out.append(x)

        if self.return_pyramid:
            out = out[3:4]

        out = self.head(out)

        return out


@MODEL_REGISTRY.register()
class PVTSegmentationV2(PVTSegmentation):
    """
    Implementation of Pyramid Vision Transformer - https://arxiv.org/abs/2102.12122v1

    Parameters
    ----------
    img_size: int
        Image size
    patch_size: list(int)
        List of patch size
    in_channels: int
        Input channels in image, default=3
    embedding_dims:  int
        Patch Embedding dimension
    num_heads:tuple[int]
        Number of heads in each transformer layer
    depths: tuple[int]
        Depth in each Transformer layer
    mlp_ratio: float
        Ratio of mlp heads to embedding dimension
    qkv_bias: bool, default= True
        Adds bias to the qkv if true
    qk_scale: float, optional
        Override default qk scale of head_dim ** -0.5 in Spatial Attention if set
    p_dropout: float,
        Dropout rate,default is 0.0
    attn_dropout:  float,
        Attention dropout rate, default is 0.0
    drop_path_rate: float
        Stochastic depth rate, default is 0.1
    sr_ratios: float
        Spatial reduction ratio
    linear: bool
        Whether to use linear spatial attention, default is False
    use_dwconv: bool
        Whether to use Depth-wise convolutions in Overlap-patch embedding, default is True
    ape: bool
        Whether to use absolute position embedding, default is False
    return_pyramid: bool
        Whether to use all pyramid feature layers for up-sampling, default is true
    """

    def __init__(
        self,
        img_size=224,
        patch_size=[7, 3, 3, 3],
        in_channels=3,
        embedding_dims=[64, 128, 256, 512],
        num_heads=[1, 2, 4, 8],
        mlp_ratio=[4, 4, 4, 4],
        qkv_bias=False,
        qk_scale=0.0,
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        ape=False,
        use_dwconv=True,
        linear=False,
        return_pyramid=False,
    ):
        super(PVTSegmentationV2, self).__init__(
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            embedding_dims=embedding_dims,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            p_dropout=p_dropout,
            attn_dropout=attn_dropout,
            drop_path_rate=drop_path_rate,
            norm_layer=norm_layer,
            depths=depths,
            sr_ratios=sr_ratios,
            linear=linear,
            ape=ape,
            use_dwconv=use_dwconv,
            return_pyramid=return_pyramid,
        )



import torch.nn as nn

from ..utils import pair


class BaseClassificationModel(nn.Module):
    """
    img_size: int
        Size of the image
    patch_size: int or tuple(int)
        Size of the patch
    in_channels: int
        Number of channels in input image
    pool: {"mean","cls"}
        Feature pooling type
    """

    def __init__(self, img_size, patch_size, in_channels=3, pool="cls"):
        super(BaseClassificationModel, self).__init__()

        img_height, img_width = pair(img_size)
        patch_height, patch_width = pair(patch_size)

        assert (
            img_height % patch_height == 0 and img_width % patch_width == 0
        ), "Image dimensions must be divisible by the patch size."

        num_patches = (img_height // patch_height) * (img_width // patch_width)
        patch_dim = in_channels * patch_height * patch_width

        self.patch_height = patch_height
        self.patch_width = patch_width
        self.num_patches = num_patches
        self.patch_dim = patch_dim

        assert pool in {
            "cls",
            "mean",
        }, "Feature pooling type must be either cls (cls token) or mean (mean pooling)"
        self.pool = pool



import torch
import torch.nn as nn
from einops import rearrange

from ..utils import ATTENTION_REGISTRY


class _Projection(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()

        if not in_dim == out_dim:
            self.l1 = nn.Linear(in_dim, out_dim)
        else:
            self.l1 = nn.Identity()

    def forward(self, x):
        return self.l1(x)


@ATTENTION_REGISTRY.register()
class CrossAttention(nn.Module):
    """
    Cross-Attention Fusion

    Parameters
    ----------
    cls_dim: int
        Dimension of cls token embedding
    patch_dim: int
        Dimension of patch token embeddings cls token to be fused with
    num_heads: int
        Number of cross-attention heads
    head_dim: int
        Dimension of each head

    """

    def __init__(self, cls_dim, patch_dim, num_heads=8, head_dim=64):
        super().__init__()

        inner_dim = num_heads * head_dim
        self.num_heads = num_heads
        self.scale = head_dim ** -0.5
        self.fl = _Projection(cls_dim, patch_dim)
        self.gl = _Projection(patch_dim, cls_dim)
        self.to_k = nn.Linear(patch_dim, inner_dim)
        self.to_v = nn.Linear(patch_dim, inner_dim)
        self.to_q = nn.Linear(patch_dim, inner_dim)
        self.cls_project = _Projection(inner_dim, patch_dim)
        self.attend = nn.Softmax(dim=-1)

    def forward(self, cls, patches):

        cls = self.fl(cls)

        x = torch.cat([cls, patches], dim=-2)
        q = self.to_q(cls)
        k = self.to_k(x)
        v = self.to_v(x)
        k = rearrange(k, "b n (h d) -> b h n d", h=self.num_heads)
        q = rearrange(q, "b n (h d) -> b h n d", h=self.num_heads)
        v = rearrange(v, "b n (h d) -> b h n d", h=self.num_heads)
        k = torch.transpose(k, -2, -1)

        attention = (q @ k) * self.scale
        attention = self.attend(attention)
        attention_value = attention @ v
        attention_value = rearrange(attention_value, "b h n d -> b n (h d)")
        attention_value = self.cls_project(attention_value)

        ycls = cls + attention_value
        ycls = self.gl(ycls)

        return ycls



import torch.nn as nn

from ..functional import PreNorm
from ..utils import ATTENTION_REGISTRY


@ATTENTION_REGISTRY.register()
class SpatialAttention(nn.Module):
    """

    Spatial Reduction Attention- Linear complexity attention layer

    Parameters
    -----------
    dim: int
        Dimension of the input tensor
    num_heads: int
        Number of attention heads
    sr_ratio :int
        Spatial Reduction ratio
    qkv_bias : bool, default is True
        If True, add a learnable bias to query, key, value.
    qk_scale : float, optional
        Override default qk scale of head_dim ** -0.5 if set
    attn_drop : float, optional
        Dropout rate
    proj_drop :float, optional
        Dropout rate
    linear : bool
        Whether to use linear Spatial attention,default is False
    act_fn : nn.Module
        Activation function, default is False

    """

    def __init__(
        self,
        dim,
        num_heads,
        sr_ratio=1,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
        linear=False,
        act_fn=nn.GELU,
    ):
        super(SpatialAttention, self).__init__()

        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.num_heads = num_heads
        self.sr_ratio = sr_ratio
        head_dim = dim // num_heads
        self.scale = qk_scale or (head_dim) ** (0.5)

        inner_dim = head_dim * num_heads

        self.q = nn.Linear(dim, inner_dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, inner_dim * 2, bias=qkv_bias)

        self.attn = nn.Sequential(nn.Softmax(dim=-1), nn.Dropout(p=attn_drop))

        self.to_out = nn.Sequential(nn.Linear(dim, dim), nn.Dropout(p=proj_drop))
        self.linear = linear
        self.sr_ratio = sr_ratio
        self.norm = PreNorm(dim=dim, fn=act_fn() if linear else nn.Identity())

        if not linear:
            if sr_ratio > 1:
                self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
        else:
            self.pool = nn.AdaptiveAvgPool2d(7)
            self.sr = nn.Conv2d(dim, dim, kernel_size=1, stride=1)

    def forward(self, x, H, W):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        H: int
            Height  of image patches
        W: int
            Width of image patches
        Returns
        ----------
        torch.Tensor
            Returns output tensor by applying spatial attention on input tensor

        """
        B, N, C = x.shape
        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if not self.linear:
            if self.sr_ratio > 1:
                x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
                x_ = self.norm(self.sr(x_).reshape(B, C, -1).permute(0, 2, 1))
                kv = (
                    self.kv(x_)
                    .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                    .permute(2, 0, 3, 1, 4)
                )
            else:
                kv = (
                    self.kv(x)
                    .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                    .permute(2, 0, 3, 1, 4)
                )
        else:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
            x_ = self.norm(self.sr(self.pool(x_)).reshape(B, C, -1).permute(0, 2, 1))
            kv = (
                self.kv(x_)
                .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )

        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = self.attn(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)

        return self.to_out(x)



from einops.layers.torch import Rearrange
from torch import nn


class LinearEmbedding(nn.Module):
    """

    Parameters
    ----------
    embedding_dim: int
        Dimension of the resultant embedding
    patch_height: int
        Height of the patch
    patch_width: int
        Width of the patch
    patch_dim: int
        Dimension of the patch

    """

    def __init__(
        self,
        embedding_dim,
        patch_height,
        patch_width,
        patch_dim,
    ):
        super().__init__()

        self.patch_embedding = nn.Sequential(
            Rearrange(
                "b c (h p1) (w p2) -> b (h w) (p1 p2 c)",
                p1=patch_height,
                p2=patch_width,
            ),
            nn.Linear(patch_dim, embedding_dim),
        )

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor

        Returns
        ----------
        torch.Tensor
            Returns patch embeddings of size `embedding_dim`

        """

        return self.patch_embedding(x)



import torch
from einops import repeat

from ...encoder import ConViTEncoder
from ...utils import MODEL_REGISTRY
from .vanilla import VanillaViT


@MODEL_REGISTRY.register()
class ConViT(VanillaViT):
    """
    Implementation of 'ConViT: Improving Vision Transformers with Soft Convolutional Inductive Biases'
    https://arxiv.org/abs/2103.10697

    Parameters
    ----------
    img_size: int
        Size of the image
    patch_size: int
        Size of a patch
    n_classes: int
        Number of classes for classification
    embedding_dim: int
        Dimension of hidden layer
    head_dim: int
        Dimension of the attention head
    depth_sa: int
        Number of attention layers in the encoder for self attention layers
    depth_gpsa: int
        Number of attention layers in the encoder for global positional self attention layers
    attn_heads_sa:int
        Number of the attention heads for self attention layers
    attn_heads_gpsa:int
        Number of the attention heads for global positional self attention layers
    encoder_mlp_dim: int
        Dimension of hidden layer in the encoder
    in_channels: int
        Number of input channels
    decoder_config: int or tuple or list, optional
        Configuration of the decoder. If None, the default configuration is used.
    pool: {"cls","mean"}
        Feature pooling type
    p_dropout_encoder: float
        Dropout probability in the encoder
    p_dropout_embedding: float
        Dropout probability in the embedding layer
    """

    def __init__(
        self,
        img_size,
        patch_size,
        n_classes,
        embedding_dim=1024,
        head_dim=64,
        depth_sa=6,
        depth_gpsa=6,
        attn_heads_sa=16,
        attn_heads_gpsa=16,
        encoder_mlp_dim=2048,
        in_channels=3,
        decoder_config=None,
        pool="cls",
        p_dropout_encoder=0,
        p_dropout_embedding=0,
    ):
        super().__init__(
            img_size,
            patch_size,
            n_classes,
            embedding_dim,
            head_dim,
            depth_sa,
            attn_heads_sa,
            encoder_mlp_dim,
            in_channels,
            decoder_config,
            pool,
            p_dropout_encoder,
            p_dropout_embedding,
        )

        self.encoder_gpsa = ConViTEncoder(
            embedding_dim,
            depth_gpsa,
            attn_heads_gpsa,
            head_dim,
            encoder_mlp_dim,
            p_dropout_encoder,
        )

    def forward(self, x):

        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of size `n_classes`

        """
        x = self.patch_embedding(x)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, "() n d -> b n d", b=b)
        x = torch.cat((cls_tokens, x), dim=1)
        x = self.pos_embedding(x)
        x_cls = x[:, 0:1, :]
        x = x[:, 1:, :]
        x = self.encoder_gpsa(x)
        x = torch.cat((x_cls, x), dim=1)
        x = self.encoder(x)
        x = self.pool(x)
        x = self.decoder(x)

        return x



import einops
import torch.nn as nn

from ...attention import VanillaSelfAttention
from ...encoder.embedding.pos_embedding import PosEmbedding
from ...utils import ATTENTION_REGISTRY, MODEL_REGISTRY


class VisformerConvBlock(nn.Module):
    """
    Convolution Block for Vision-Friendly transformers
    https://arxiv.org/abs/2104.12533

    Parameters
    ----------
    in_channels: int
        Number of input channels
    group: int
        Number of groups for convolution, default is 8
    activation: torch.nn.Module
        Activation function between layers, default is nn.GELU
    p_dropout: float
        Dropout rate, default is 0.0
    """

    def __init__(self, in_channels, group=8, activation=nn.GELU, p_dropout=0.0):
        super().__init__()

        self.norm1 = nn.BatchNorm2d(in_channels)
        self.conv1 = nn.Conv2d(in_channels, in_channels * 2, kernel_size=1, bias=False)
        self.act1 = activation()
        self.conv2 = nn.Conv2d(
            in_channels * 2,
            in_channels * 2,
            kernel_size=3,
            padding=1,
            groups=group,
            bias=False,
        )
        self.act2 = activation()
        self.conv3 = nn.Conv2d(in_channels * 2, in_channels, kernel_size=1, bias=False)
        self.drop = nn.Dropout(p_dropout)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of same size as input
        """

        xt = x
        xt = self.norm1(xt)
        x = self.conv1(x)
        x = self.act1(x)
        x = self.drop(x)
        x = self.conv2(x)
        x = self.act2(x)
        x = self.conv3(x)
        x = self.drop(x)
        x = x + xt

        return x


@ATTENTION_REGISTRY.register()
class VisformerAttentionBlock(nn.Module):
    """
    Attention Block for Vision-Friendly transformers
    https://arxiv.org/abs/2104.12533

    Parameters
    ----------
    in_channels: int
        Number of input channels
    num_heads: int
        Number of heads for attention, default is 8
    activation: torch.nn.Module
        Activation function between layers, default is nn.GELU
    p_dropout: float
        Dropout rate, default is 0.0
    """

    def __init__(self, in_channels, num_heads=8, activation=nn.GELU, p_dropout=0.0):
        super().__init__()

        self.conv1 = nn.Conv2d(in_channels, in_channels * 4, kernel_size=1, bias=False)
        self.act1 = activation()
        self.conv2 = nn.Conv2d(in_channels * 4, in_channels, kernel_size=1, bias=False)
        self.attn = VanillaSelfAttention(
            in_channels, num_heads=num_heads, head_dim=in_channels // num_heads
        )
        self.norm1 = nn.BatchNorm2d(in_channels)
        self.norm2 = nn.BatchNorm2d(in_channels)
        self.drop = nn.Dropout(p_dropout)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of same size as input
        """

        B, C, H, W = x.shape
        xt = einops.rearrange(x, "b c h w -> b (h w) c")
        x = self.norm1(x)
        x = einops.rearrange(x, "b c h w -> b (h w) c")
        x = self.attn(x)
        x = xt + x
        x = einops.rearrange(x, "b (h w) c -> b c h w", h=H, w=W)
        xt = x
        x = self.norm2(x)
        x = self.conv1(x)
        x = self.act1(x)
        x = self.conv2(x)
        x = self.drop(x)
        x = xt + x

        return x


@MODEL_REGISTRY.register()
class Visformer(nn.Module):
    """
    A builder to construct a Vision-Friendly transformer model as in the paper: "Visformer: A vision-friendly transformer"
    https://arxiv.org/abs/2104.12533

    Parameters
    ----------
    img_size: int,tuple
        Size of the input image
    n_classes: int
        Number of classes in the dataset
    depth: tuple[int]
        Number of layers before each embedding reduction
    config: tuple[int]
        Choice of convolution block (0) or attention block (1) for corresponding layer
    channel_config: tuple[int]
        Number of channels for each layer
    num_heads: int
        Number of heads for attention block, default is 8
    conv_group: int
        Number of groups for convolution block, default is 8
    p_dropout_conv: float
        Dropout rate for convolution block, default is 0.0
    p_dropout_attn: float
        Dropout rate for attention block, default is 0.0
    activation: torch.nn.Module
        Activation function between layers, default is nn.GELU
    pos_embedding: bool
        Whether to use positional embedding, default is True

    """

    def __init__(
        self,
        img_size,
        n_classes,
        depth: tuple,
        config: tuple,
        channel_config: tuple,
        num_heads=8,
        conv_group=8,
        p_dropout_conv=0.0,
        p_dropout_attn=0.0,
        activation=nn.GELU,
        pos_embedding=True,
    ):
        super().__init__()

        q = 0
        assert (
            len(channel_config) == len(depth) - depth.count(0) + 2
        ), "Channel config is not correct"

        assert set(config).issubset(
            set([0, 1])
        ), "Config is not correct, should contain only 0 and 1"

        self.linear = nn.Linear(channel_config[-1], n_classes)

        if isinstance(img_size, int):
            img_size = (img_size, img_size)
        image_size = list(img_size)

        assert image_size[0] // (2 ** (len(depth) + 1)) > 0, "Image size is too small"
        assert image_size[1] // (2 ** (len(depth) + 1)) > 0, "Image size is too small"

        self.stem = nn.ModuleList(
            [
                nn.Conv2d(
                    channel_config[q],
                    channel_config[q + 1],
                    kernel_size=7,
                    padding=3,
                    stride=2,
                    bias=False,
                ),
                nn.BatchNorm2d(channel_config[q + 1]),
                nn.ReLU(inplace=True),
            ]
        )

        q += 1
        emb = 2
        image_size = [i // 2 for i in image_size]

        for i in range(len(depth)):

            if depth[i] == 0:
                emb *= 2
                config = tuple([0] + list(config))
                continue

            self.stem.extend(
                [
                    nn.Conv2d(
                        channel_config[q],
                        channel_config[q + 1],
                        kernel_size=emb,
                        stride=emb,
                    ),
                    nn.BatchNorm2d(channel_config[q + 1]),
                    nn.ReLU(inplace=True),
                ]
            )

            image_size = [k // emb for k in image_size]
            emb = 2
            q += 1

            if pos_embedding:
                self.stem.extend(
                    [PosEmbedding([channel_config[q], image_size[0]], image_size[1])]
                )

            if config[i] == 0:
                self.stem.extend(
                    [
                        VisformerConvBlock(
                            channel_config[q],
                            group=conv_group,
                            p_dropout=p_dropout_conv,
                            activation=activation,
                        )
                        for j in range(depth[i])
                    ]
                )

            elif config[i] == 1:
                self.stem.extend(
                    [
                        VisformerAttentionBlock(
                            channel_config[q],
                            num_heads,
                            activation,
                            p_dropout_attn,
                        )
                        for j in range(depth[i])
                    ]
                )

        self.stem.extend([nn.BatchNorm2d(channel_config[-1]), nn.AdaptiveAvgPool2d(1)])
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns tensor of size `n_classes`

        """
        for i in self.stem:
            x = i(x)

        x.squeeze_(2).squeeze_(2)
        x = self.linear(x)
        x = self.softmax(x)

        return x


@MODEL_REGISTRY.register()
def Visformer_S(img_size, n_classes, in_channels=3):
    """
    Visformer-S model from the paper:"Visformer: The Vision-friendly Transformer"
    https://arxiv.org/abs/1906.11488

    Parameters
    ----------
    img_size: int,tuple
        Size of the input image
    n_classes: int
        Number of classes in the dataset
    in_channels: int
        Number of channels in the input
    """
    return Visformer(
        img_size,
        n_classes,
        (0, 7, 4, 4),
        (0, 1, 1),
        (in_channels, 32, 192, 384, 768),
        num_heads=6,
    )


@MODEL_REGISTRY.register()
def VisformerV2_S(img_size, n_classes, in_channels=3):
    """
    VisformerV2-S model from the paper:"Visformer: The Vision-friendly Transformer"
    https://arxiv.org/abs/1906.11488

    Parameters
    ----------
    img_size: int,tuple
        Size of the input image
    n_classes: int
        Number of classes in the dataset
    in_channels: int
        Number of channels in the input
    """
    return Visformer(
        img_size,
        n_classes,
        (1, 10, 14, 3),
        (0, 0, 1, 1),
        (in_channels, 32, 64, 128, 256, 512),
        num_heads=6,
    )


@MODEL_REGISTRY.register()
def Visformer_Ti(img_size, n_classes, in_channels=3):
    """
    Visformer-Ti model from the paper:"Visformer: The Vision-friendly Transformer"
    https://arxiv.org/abs/1906.11488

    Parameters
    ----------
    img_size: int,tuple
        Size of the input image
    n_classes: int
        Number of classes in the dataset
    in_channels: int
        Number of channels in the input
    """
    return Visformer(
        img_size,
        n_classes,
        (0, 7, 4, 4),
        (0, 1, 1),
        (in_channels, 16, 96, 192, 384),
        num_heads=6,
    )


@MODEL_REGISTRY.register()
def VisformerV2_Ti(img_size, n_classes, in_channels=3):
    """
    VisformerV2-Ti model from the paper:"Visformer: The Vision-friendly Transformer"
    https://arxiv.org/abs/1906.11488

    Parameters
    ----------
    img_size: int,tuple
        Size of the input image
    n_classes: int
        Number of classes in the dataset
    in_channels: int
        Number of channels in the input
    """

    return Visformer(
        img_size,
        n_classes,
        (1, 4, 6, 2),
        (0, 0, 1, 1),
        (in_channels, 24, 48, 96, 192, 384),
        num_heads=6,
    )



import codecs
import os

from setuptools import find_packages, setup

# Basic information
NAME = "vformer"
DESCRIPTION = "A modular PyTorch library for vision transformer models"
VERSION = "0.1.1"
AUTHOR = "Neelay Shah"
EMAIL = "nstraum1@gmail.com"
LICENSE = "MIT"
REPOSITORY = "https://github.com/SforAiDl/vformer"
PACKAGE = "SforAiDl"

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

# Define the keywords
KEYWORDS = [
    "vision transformers",
    "pytorch",
    "computer vision",
    "machine learning",
    "deep learning",
]

# Define the classifiers
# See https://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]

# Important Paths
PROJECT = os.path.abspath(os.path.dirname(__file__))
REQUIRE_PATH = "requirements.txt"
PKG_DESCRIBE = "README.md"

# Directories to ignore in find_packages
EXCLUDES = ()


# helper functions
def read(*parts):
    """
    returns contents of file
    """
    with codecs.open(os.path.join(PROJECT, *parts), "rb", "utf-8") as file:
        return file.read()


def get_requires(path=REQUIRE_PATH):
    """
    generates requirements from file path given as REQUIRE_PATH
    """
    for line in read(path).splitlines():
        line = line.strip()
        if line and not line.startswith("#"):
            yield line


# Define the configuration
CONFIG = {
    "name": NAME,
    "version": VERSION,
    "description": DESCRIPTION,
    "long_description": LONG_DESCRIPTION,
    "long_description_content_type": "text/markdown",
    "classifiers": CLASSIFIERS,
    "keywords": KEYWORDS,
    "license": LICENSE,
    "author": AUTHOR,
    "author_email": EMAIL,
    "url": REPOSITORY,
    "project_urls": {"Source": REPOSITORY},
    "packages": find_packages(
        where=PROJECT, include=["vformer", "vformer.*"], exclude=EXCLUDES
    ),
    "install_requires": list(get_requires()),
    "python_requires": ">=3.6",
    "test_suite": "tests",
    "tests_require": ["pytest>=3"],
    "include_package_data": True,
}

if __name__ == "__main__":
    setup(**CONFIG)



import torch
import torch.nn as nn
from einops import rearrange

from ..utils import ATTENTION_REGISTRY


@ATTENTION_REGISTRY.register()
class VanillaSelfAttention(nn.Module):
    """
    Vanilla O(n^2) Self attention

    Parameters
    ----------
    dim: int
        Dimension of the embedding
    num_heads: int
        Number of the attention heads
    head_dim: int
        Dimension of each head
    p_dropout: float
        Dropout Probability

    """

    def __init__(self, dim, num_heads=8, head_dim=64, p_dropout=0.0):
        super().__init__()

        inner_dim = head_dim * num_heads
        project_out = not (num_heads == 1 and head_dim == dim)

        self.num_heads = num_heads
        self.scale = head_dim ** -0.5

        self.attend = nn.Softmax(dim=-1)
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)

        self.to_out = (
            nn.Sequential(nn.Linear(inner_dim, dim), nn.Dropout(p_dropout))
            if project_out
            else nn.Identity()
        )

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor
        Returns
        ----------
        torch.Tensor
            Returns output tensor by applying self-attention on input tensor

        """
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.num_heads), qkv
        )

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)

        out = torch.matmul(attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")

        return self.to_out(out)



from .cct import CCT
from .convit import ConViT
from .convvt import ConvVT
from .cross import CrossViT
from .cvt import CVT
from .pyramid import PVTClassification, PVTClassificationV2
from .swin import SwinTransformer
from .vanilla import VanillaViT
from .visformer import *
from .vivit import ViViTModel2, ViViTModel3



from .convvt import ConvVTAttention
from .cross import CrossAttention
from .gated_positional import GatedPositionalSelfAttention
from .spatial import SpatialAttention
from .vanilla import VanillaSelfAttention
from .window import WindowAttention
